/**
 *	@file NPWBase.c
 *	@brief NPW (base functions to write PDF files)
 *	@author Yves Piguet.
 *	@copyright 2008-2014, Yves Piguet
 */

/* License: new BSD license (see header file) */

#include "NPWBase.h"

#if !defined(NPWNoFloat)
#	include <math.h>
#endif

/// PDF header string
#define kHeaderStr "%PDF-1.5\n"

/// Check error returned by a function and returns it unless kNPWErrOk
#define Chk(e) \
	do { \
		err = (e); \
		if (err != kNPWErrOk) \
			return err; \
	} while (0)

/// Check error returned by a function and goto error unless kNPWErrOk
#define ChkG(e) \
	do { \
		err = (e); \
		if (err != kNPWErrOk) \
			goto error; \
	} while (0)

#define kDigits 8	///< maximum number of digits

/// Convert a value in [0,15] to its ASCII hexadecimal representation
#define hex(n) ((n) >= 10 ? 'a' - 10 + (n) : '0' + (n))

static NPWConstText kDataKeyPage = "page";
static NPWConstText kDataKeyXRef = "xref";
static NPWConstText kDataKeyRsrcCB = "rsrcCB";
static NPWConstText kDataKeyDelayedActionCB = "delayedActionCB";

void NPWInit(NPW *pdf)
{
	pdf->fonts = kNPWFontAll;
	pdf->offset = 0;
	pdf->inPage = FALSE;
	pdf->writingObject = FALSE;
	pdf->alloc = NULL;
	pdf->write = NULL;
	pdf->data = NULL;
	pdf->dataSize = pdf->dataAllocatedSize = 0;
	pdf->xrefCount = 0;
	pdf->pageCount = 0;
	pdf->info = kNPWNoRef;
	pdf->filter = kNPWFilterNone;
	pdf->asciiBufCount = pdf->asciiFilterLineLength = 0;
}

void NPWSetAllocCallback(NPW *pdf,
		NPWAllocCB alloc, void *userData)
{
	pdf->alloc = alloc;
	pdf->allocUserData = userData;
}

void NPWSetWriteCallback(NPW *pdf,
		NPWWriteCB write, void *userData)
{
	pdf->write = write;
	pdf->writeUserData = userData;
}

void NPWSetFonts(NPW *pdf, NPWInt fonts)
{
	pdf->fonts = fonts;
}

NPWErr NPWWrite(NPW *pdf, NPWConstText str, NPWInt length)
{
	NPWErr err = kNPWErrOk;
	
	if (length < 0)
		for (length = 0; str[length] != '\0'; length++)
			;
	if (pdf->write)
		err = pdf->write(str, length, pdf->writeUserData);
	pdf->offset += length;
	return err;
}

NPWErr NPWWriteInt(NPW *pdf, NPWInt i, NPWInt pad)
{
	NPWChar str[20];
	NPWInt len = 0;
	NPWInt j;
	
	if (i == 0 && pad < 0)
		str[len++] = '0';
	else
	{
		if (i < 0)
		{
			str[len++] = '-';
			i = -i;
			pad--;
		}
		if (pad < 0)
		{
			// count len
			for (j = i; j > 0; j /= 10, len++)
				;
			// output digits
			for (j = 1; i > 0; i /= 10, j++)
				str[len - j] = '0' + i % 10;
		}
		else
		{
			len += pad;
			for (j = 1; j <= pad; i /= 10, j++)
				str[len - j] = '0' + i % 10;
		}
	}
	
	return NPWWrite(pdf, str, len);
}

#if !defined(NPWNoFloat)
NPWErr NPWWriteFloat(NPW *pdf, NPWFloat x)
{
	NPWInt n, i;
	NPWFloat pw10;
	NPWChar d;
	NPWChar frac[kDigits];
	NPWErr err;
	
	// minus sign
	if (x < 0)
	{
		Chk(NPWWrite(pdf, "-", 1));
		x = -x;
	}
	
	// inf
	if (x > 1e30)
		return kNPWNumTooLarge;
	
	// integer digits
	for (pw10 = 1, n = 1; 10 * pw10 <= x; pw10 *= 10, n++)
		;
	do
	{
		d = '0' + (int)fmod(floor(x / pw10), 10);
		Chk(NPWWrite(pdf, &d, 1));
		pw10 /= 10;
	} while (pw10 > 0.5);
	
	// fractional part
	for (i = 0, pw10 = 10; n + i < kDigits; i++, pw10 *= 10)
		frac[i] = '0' + (int)fmod(floor(x * pw10), 10);
	while (i > 0 && frac[i - 1] == '0')
		i--;
	if (i > 0)	// else integer (always legal in PDF)
	{
		Chk(NPWWrite(pdf, ".", 1));
		Chk(NPWWrite(pdf, frac, i));
	}
	return kNPWErrOk;
}
#endif

NPWErr NPWWriteString(NPW *pdf, NPWConstText str, NPWInt length)
{
	NPWInt i, j;
	NPWChar str1[8];
	NPWErr err;
	
	Chk(NPWWrite(pdf, "(", 1));
	for (i = 0; length < 0 ? str[i] : i < length; )
	{
		for (j = 0;
				(length < 0 || i + j < length)
					&& (str[i + j] < 0
						|| (str[i + j] >= ' ' && str[i + j] != '(' && str[i + j] != ')'));
				j++)
			;
		if (j > 0)
			Chk(NPWWrite(pdf, str + i, j));
		i += j;
		if (length < 0 ? str[i] : i < length)
		{
			str1[0] = '\\';
			j = 2;
			switch (str[i])
			{
				case 10:
					str1[1] = 'n';
					break;
				case 13:
					str1[1] = 'r';
					break;
				case 9:
					str1[1] = 't';
					break;
				case 8:
					str1[1] = 'b';
					break;
				case '(':
					str1[1] = '(';
					break;
				case ')':
					str1[1] = ')';
					break;
				default:
					str1[1] = '0' + (0777 & str[i]) / 0100;
					str1[2] = '0' + (077 & str[i]) / 010;
					str1[3] = '0' + (07 & str[i]);
					j = 4;
					break;
			}
			Chk(NPWWrite(pdf, str1, j));
			i++;
		}
	}
	Chk(NPWWrite(pdf, ")", 1));
	return kNPWErrOk;
}

/// State of run-length filter
typedef struct
{
	NPWInt offset;	///< offset in input buffer
	NPWBoolean done;	///< TRUE once the last byte of encoded data has been output
	NPWBoolean run;	///< TRUE if next byte is the value of a run
	NPWByte val;	///< value of a run
	NPWInt rem;	///< number of remaining bytes in a run of different values
} RLState;

#define kRLEOD 128	///< end of data

/**	Initialize run-length filter state
	@param[out] rl run-length filter state
*/
static void RLInit(RLState *rl)
{
	rl->offset = 0;
	rl->done = FALSE;
	rl->run = FALSE;
	rl->val = 0;
	rl->rem = 0;
}

/**	Get next output byte of run-length filtered stream
	@param[in,out] rl run-length filter state
	@param[in] b input data
	@param[in] len number of bytes in b
	@return output byte
*/
static NPWByte RLNextByte(RLState *rl, NPWByte const *b, NPWInt len)
{
	if (rl->run)
	{
		rl->run = FALSE;
		rl->rem = 0;
		return rl->val;
	}
	else if (rl->rem > 0)
	{
		rl->rem--;
		return b[rl->offset++];
	}
	else if (rl->offset >= len)
	{
		rl->done = TRUE;
		return kRLEOD;
	}
	else
	{
		// next run
		NPWInt n;
		
		// count indentical bytes
		for (n = 1;
				n < 128 && rl->offset + n < len && b[rl->offset + n] == b[rl->offset];
				n++)
			;
		if (n >= 2)
		{
			rl->val = b[rl->offset];
			rl->offset += n;
			rl->run = TRUE;
			return (NPWByte)(257 - n);
		}
		
		// count different bytes
		for (n = 1;
				n < 128 && rl->offset + n < len
						&& b[rl->offset + n] != b[rl->offset + n - 1];
				n++)
			;
		if (rl->offset + n < len && b[rl->offset + n - 1] == b[rl->offset + n])
			n--;	// last byte belong to next run
		rl->rem = n;
		return (NPWByte)(n - 1);
	}
}

void NPWFilterBegin(NPW *pdf, NPWInt filter)
{
	pdf->filter = filter;
	pdf->asciiBufCount = 0;
	pdf->asciiFilterLineLength = 0;
}

/**	Write 4 bytes, or less if last bytes of stream, encoded with base-85
	@param pdf reference to NPW structure
	@param[in] b bytes
	@param[in] n number of bytes (1-4)
	@param[in] end TRUE to write the end marker
	@return error code (kNPWErrOk for success)
*/
static NPWErr writeA85(NPW *pdf,
		NPWByte const b[/* 4 */], NPWInt n, NPWBoolean end)
{
	NPWInt i;
	NPWChar s[5];
	NPWUInt32 x;
	NPWErr err;
	
	for (i = 0, x = 0; i < n; i++)
		x = (x << 8) | b[i];
	if (n < 4)
		x <<= 8 * (4 - n);
	for (i = 0; i < 5; i++, x /= 85)
		s[4 - i] = '!' + x % 85;
	if (n == 4
			&& s[0] == '!' && s[1] == '!' && s[2] == '!' && s[3] == '!' && s[4] == '!')
		err = NPWWrite(pdf, "z", 1);
	else
		err = NPWWrite(pdf, s, 1 + n);
	if (err == kNPWErrOk && end)
		err = NPWWrite(pdf, "~>\n", -1);
	return err;
}

/**	Write a byte as ASCII hexadecimal
	@param pdf reference to NPW structure
	@param[in] b byte
	@return error code (kNPWErrOk for success)
*/
static NPWErr writeAHex(NPW *pdf, NPWByte b)
{
	NPWChar s[2];
	
	s[0] = hex(b >> 4);
	s[1] = hex(b & 0xf);
	return NPWWrite(pdf, s, 2);
}

NPWErr NPWFilterWrite(NPW *pdf, NPWByte const *b, NPWInt n)
{
	NPWInt i = 0;
	RLState rl;
	NPWErr err;
	
	if (pdf->filter & kNPWFilterRunLength)
		RLInit(&rl);
	
	// write sequences of 4 bytes
	if (pdf->filter & kNPWFilterRunLength)
	{
		if (pdf->filter & kNPWFilterASCII85)
			while (!rl.done)
			{
				pdf->asciiBuf[pdf->asciiBufCount++] = RLNextByte(&rl, b, n);
				// write output if filled
				if (pdf->asciiBufCount == 4)
				{
					Chk(writeA85(pdf, pdf->asciiBuf, 4, FALSE));
					pdf->asciiBufCount = 0;
				}
			}
		else if (pdf->filter & kNPWFilterASCIIHex)
			while (!rl.done)
				Chk(writeAHex(pdf, RLNextByte(&rl, b, n)));
		else
			while (!rl.done)
			{
				// get and write one byte
				pdf->asciiBuf[0] = RLNextByte(&rl, b, n);
				Chk(NPWWrite(pdf, (NPWConstText)pdf->asciiBuf, 1));
			}
	}
	else
		if (pdf->filter & kNPWFilterASCII85)
		{
			for (; i < n; i++)
			{
				pdf->asciiBuf[pdf->asciiBufCount++] = b[i];
				// write output if filled
				if (pdf->asciiBufCount == 4)
				{
					Chk(writeA85(pdf, pdf->asciiBuf, 4, FALSE));
					pdf->asciiBufCount = 0;
				}
			}
		}
		else if (pdf->filter & kNPWFilterASCIIHex)
			for (; i < n; i++)
				Chk(writeAHex(pdf, b[i]));
		else
			Chk(NPWWrite(pdf, (NPWConstText)b, n));
	
	return kNPWErrOk;
}

NPWErr NPWFilterEnd(NPW *pdf)
{
	if (pdf->filter & kNPWFilterASCII85)
		return writeA85(pdf, pdf->asciiBuf, pdf->asciiBufCount, TRUE);
	else
		return NPWWrite(pdf, "\n", 1);
}

/** Add an entry to the cross-reference table.
	@param pdf reference to NPW structure
*/
static NPWErr addXRef(NPW *pdf)
{
	NPWXRef *xref;
	NPWErr err;
	
	Chk(NPWDataAlloc(pdf, kDataKeyXRef, sizeof(NPWXRef), (void **)&xref));
	xref->offset = pdf->offset;
	xref->generation = 0;
	pdf->xrefCount++;
	return kNPWErrOk;
}

/** Reserve an entry in the cross-reference table.
	@param pdf reference to NPW structure
	@param[out] ref object reference
*/
static NPWErr reserveXRef(NPW *pdf, NPWRef *ref)
{
	NPWXRef *xref;
	NPWErr err;
	
	Chk(NPWDataAlloc(pdf, kDataKeyXRef, sizeof(NPWXRef), (void **)&xref));
	xref->offset = -1;
	xref->generation = -1;
	pdf->xrefCount++;
	if (ref)
		*ref = pdf->xrefCount;
	return kNPWErrOk;
}

/** Write PDF header.
	@param pdf reference to NPW structure
	@return kNPWErrOk for success, else error code
*/
static NPWErr writeHeader(NPW *pdf)
{
	return NPWWrite(pdf, kHeaderStr, -1);
}

NPWErr NPWBeginObj(NPW *pdf, NPWRef *ref)
{
	NPWErr err;
	
	if (pdf->writingObject)
		return kNPWErrNestedObject;
	
	Chk(addXRef(pdf));
	Chk(NPWWriteInt(pdf, pdf->xrefCount, -1));
	Chk(NPWWrite(pdf, " 0 obj\n", -1));
	if (ref)
		*ref = pdf->xrefCount;
	pdf->writingObject = TRUE;
	return kNPWErrOk;
}

/**	Write the beginning of an object with the last xref obtained
	with reserveXRef.
	@param pdf reference to NPW structure
	@param[out] ref 
	@return kNPWErrOk for success, else error code
*/
static NPWErr beginReservedObj(NPW *pdf, NPWRef *ref)
{
	NPWXRef *xref;
	NPWInt i;
	NPWErr err;
	
	if (pdf->writingObject)
		return kNPWErrNestedObject;
	
	for (i = pdf->xrefCount - 1; i >= 0; i--)
	{
		Chk(NPWDataGetNth(pdf, kDataKeyXRef, i, (void **)&xref));
		if (xref->offset < 0)
			break;
	}
	if (i < 0)
		return kNPWErrInternal;
	xref->offset = pdf->offset;
	xref->generation = 0;
	Chk(NPWWriteInt(pdf, i + 1, -1));
	Chk(NPWWrite(pdf, " 0 obj\n", -1));
	if (ref)
		*ref = i + 1;
	pdf->writingObject = TRUE;
	return kNPWErrOk;
}

NPWErr NPWEndObj(NPW *pdf)
{
	if (!pdf->writingObject)
		return kNPWErrUnbalancedObj;
	pdf->writingObject = FALSE;
	return NPWWrite(pdf, "endobj\n\n", -1);
}

static NPWErr writePage(NPW *pdf, NPWInt i, NPWRef parent, NPWRef *ref)
{
	NPWPage *page0, *page;
	NPWBoolean oneSize;	///< TRUE iff all pages have the same size inherited from Pages
	NPWErr err;
	
	page0 = NULL;
	Chk(NPWDataGetNext(pdf, kDataKeyPage, (void **)&page0));
	for (page = page0, oneSize = TRUE; ; )
	{
		if (NPWDataGetNext(pdf, kDataKeyPage, (void **)&page) != kNPWErrOk)
			break;
		oneSize &= page->width == page0->width
				&& page->height == page0->height;
	}
	Chk(NPWDataGetNth(pdf, kDataKeyPage, i, (void **)&page));
	
	Chk(NPWBeginObj(pdf, ref));
	Chk(NPWWrite(pdf, "<<\n/Type /Page\n/Parent ", -1));
	Chk(NPWWriteInt(pdf, parent, -1));
	Chk(NPWWrite(pdf, " 0 R\n", -1));
	if (page->contents >= 0)
	{
		Chk(NPWWrite(pdf, "/Contents ", -1));
		Chk(NPWWriteInt(pdf, page->contents, -1));
		Chk(NPWWrite(pdf, " 0 R\n", -1));
	}
	if (!oneSize)
	{
		Chk(NPWWrite(pdf, "/MediaBox [0 0 ", -1));
		Chk(NPWWriteInt(pdf, page->width, -1));
		Chk(NPWWrite(pdf, " ", -1));
		Chk(NPWWriteInt(pdf, page->height, -1));
		Chk(NPWWrite(pdf, "]\n", -1));
	}
	Chk(NPWWrite(pdf, ">>\n", -1));
	Chk(NPWEndObj(pdf));
	return kNPWErrOk;
}

static NPWErr writePages(NPW *pdf, NPWRef *ref)
{
	NPWPage *page0, *page;
	NPWInt i;
	NPWBoolean oneSize;	///< TRUE iff all pages have the same size
	NPWActionCBData *rsrcCBData;
	NPWErr err;
	
	Chk(NPWDataGetNth(pdf, kDataKeyPage, 0, (void **)&page0));
	
	Chk(NPWBeginObj(pdf, ref));
	Chk(NPWWrite(pdf, "<<\n/Type /Pages\n/Kids\n[\n", -1));
	for (i = 0, oneSize = TRUE, page = NULL; i < pdf->pageCount; i++)
	{
		Chk(NPWDataGetNext(pdf, kDataKeyPage, (void **)&page));
		Chk(NPWWriteInt(pdf, *ref + 1 + i, -1));
		Chk(NPWWrite(pdf, " 0 R\n", -1));
		oneSize &= page->width == page0->width
				&& page->height == page0->height;
	}
	Chk(NPWWrite(pdf, "]\n/Count ", -1));
	Chk(NPWWriteInt(pdf, pdf->pageCount, -1));
	Chk(NPWWrite(pdf, "\n", -1));
	if (pdf->pageCount > 0 && oneSize)
	{
		Chk(NPWWrite(pdf, "/MediaBox [0 0 ", -1));
		Chk(NPWWriteInt(pdf, page0->width, -1));
		Chk(NPWWrite(pdf, " ", -1));
		Chk(NPWWriteInt(pdf, page0->height, -1));
		Chk(NPWWrite(pdf, "]\n", -1));
	}
	Chk(NPWWrite(pdf, "/Resources\n<<\n", -1));
	if (pdf->fonts != kNPWFontNone)
	{
		Chk(NPWWrite(pdf, "/Font\n<<\n", -1));
		if (pdf->fonts & kNPWFontCourier)
			Chk(NPWWrite(pdf,
					NPWFontNameCourier
						" << /Type /Font /Subtype /Type1 /BaseFont /Courier /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontCourierOblique)
			Chk(NPWWrite(pdf,
					NPWFontNameCourierOblique
						" << /Type /Font /Subtype /Type1 /BaseFont /Courier-Oblique /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontCourierBold)
			Chk(NPWWrite(pdf,
					NPWFontNameCourierBold
						" << /Type /Font /Subtype /Type1 /BaseFont /Courier-Bold /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontCourierBoldOblique)
			Chk(NPWWrite(pdf,
					NPWFontNameCourierBoldOblique
						" << /Type /Font /Subtype /Type1 /BaseFont /Courier-BoldOblique /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontHelvetica)
			Chk(NPWWrite(pdf,
					NPWFontNameHelvetica
						" << /Type /Font /Subtype /Type1 /BaseFont /Helvetica /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontHelveticaOblique)
			Chk(NPWWrite(pdf,
					NPWFontNameHelveticaOblique
						" << /Type /Font /Subtype /Type1 /BaseFont /Helvetica-Oblique /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontHelveticaBold)
			Chk(NPWWrite(pdf,
					NPWFontNameHelveticaBold
						" << /Type /Font /Subtype /Type1 /BaseFont /Helvetica-Bold /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontHelveticaBoldOblique)
			Chk(NPWWrite(pdf,
					NPWFontNameHelveticaBoldOblique
						" << /Type /Font /Subtype /Type1 /BaseFont /Helvetica-BoldOblique /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontTimesRoman)
			Chk(NPWWrite(pdf,
					NPWFontNameTimesRoman
						" << /Type /Font /Subtype /Type1 /BaseFont /Times-Roman /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontTimesItalic)
			Chk(NPWWrite(pdf,
					NPWFontNameTimesItalic
						" << /Type /Font /Subtype /Type1 /BaseFont /Times-Italic /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontTimesBold)
			Chk(NPWWrite(pdf,
					NPWFontNameTimesBold
						" << /Type /Font /Subtype /Type1 /BaseFont /Times-Bold /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontTimesBoldItalic)
			Chk(NPWWrite(pdf,
					NPWFontNameTimesBoldItalic
						" << /Type /Font /Subtype /Type1 /BaseFont /Times-BoldItalic /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontSymbol)
			Chk(NPWWrite(pdf,
					NPWFontNameSymbol
						" << /Type /Font /Subtype /Type1 /BaseFont /Symbol /Encoding /WinAnsiEncoding >>\n",
					-1));
		if (pdf->fonts & kNPWFontZapfDingbats)
			Chk(NPWWrite(pdf,
					NPWFontNameZapfDingbats
						" << /Type /Font /Subtype /Type1 /BaseFont /ZapfDingbats /Encoding /WinAnsiEncoding >>\n",
					-1));
		Chk(NPWWrite(pdf, ">>\n", -1));	// Font
	}
	rsrcCBData = NULL;
	for (;;)
	{
		if (NPWDataGetNext(pdf, kDataKeyRsrcCB, (void **)&rsrcCBData) != kNPWErrOk)
			break;
		Chk(rsrcCBData->cb(pdf, rsrcCBData->userData));
	}
	Chk(NPWWrite(pdf, ">>\n", -1));	// Resources
	Chk(NPWWrite(pdf, ">>\n", -1));	// Pages
	Chk(NPWEndObj(pdf));
	
	return kNPWErrOk;
}

/**	Write PDF catalog.
	@param pdf reference to NPW structure
	@param[in] pages reference of the Pages object
	@param[out] ref reference of the PDF catalog
	@return error code (kNPWErrOk for success)
*/
static NPWErr writeCatalog(NPW *pdf, NPWRef pages, NPWRef *ref)
{
	NPWErr err;
	
	Chk(NPWBeginObj(pdf, ref));
	Chk(NPWWrite(pdf, "<<\n/Type /Catalog\n/Pages ", -1));
	Chk(NPWWriteInt(pdf, pages, -1));
	Chk(NPWWrite(pdf, " 0 R\n>>\n", -1));
	Chk(NPWEndObj(pdf));
	return kNPWErrOk;
}

/// Header of pieces of data
typedef struct
{
	NPWConstText key;	///< key used to identify the type of data
	NPWInt payloadSize;	///< size of data following the header, including padding
} NPWDataHeader;

#define kSizeofDataHeaderRoundedUp (2 * sizeof(void *))

NPWErr NPWDataAlloc(NPW *pdf, NPWConstText key, NPWInt size, void **addr)
{
	NPWDataHeader *header;
	NPWErr err;
	
	// add padding to enforce correct alignment
	size = size + (kSizeofDataHeaderRoundedUp - 1) & ~(kSizeofDataHeaderRoundedUp - 1);
	
	// make sure there is enough space in data[]
	if (sizeof(NPWDataHeader) + size > pdf->dataAllocatedSize - pdf->dataSize)
	{
		// must expand data
		if (!pdf->alloc)
			return kNPWErrNotEnoughMemory;
		if (pdf->dataAllocatedSize > 2 * (sizeof(NPWDataHeader) + size))
			pdf->dataAllocatedSize *= 2;
		else
			pdf->dataAllocatedSize = 4 * (sizeof(NPWDataHeader) + size);
		Chk(pdf->alloc(&pdf->data,
				pdf->dataAllocatedSize,
				pdf->allocUserData));
	}
	
	header = (NPWDataHeader *)((NPWByte *)pdf->data + pdf->dataSize);
	header->key = key;
	header->payloadSize = size;
	pdf->dataSize += sizeof(NPWDataHeader) + size;
	*addr = (void *)&header[1];
	
	return kNPWErrOk;
}

NPWInt NPWDataCount(NPW *pdf, NPWConstText key)
{
	void *headerAddr;
	NPWInt n = 0;
	
	if (!pdf->data)
		return 0;
	for (headerAddr = pdf->data;
			headerAddr < (void *)((NPWByte *)pdf->data + pdf->dataSize);
			headerAddr = (void *)((NPWByte *)headerAddr + sizeof(NPWDataHeader)
					+ ((NPWDataHeader const *)headerAddr)->payloadSize))
		if (((NPWDataHeader const *)headerAddr)->key == key)
			n++;
	
	return n;
}

NPWErr NPWDataGetNext(NPW *pdf, NPWConstText key, void **addr)
{
	if (!pdf->data)
		return kNPWErrDataNotFound;
	if (!*addr)
		*addr = (void *)((NPWByte *)pdf->data + sizeof(NPWDataHeader));
	else
		*addr = (void *)((NPWByte *)*addr + sizeof(NPWDataHeader)
				+ ((NPWDataHeader const *)*addr)[-1].payloadSize);
	
	for (;
			*addr < (void *)((NPWByte *)pdf->data + pdf->dataSize);
			*addr = (void *)((NPWByte *)*addr + sizeof(NPWDataHeader)
					+ ((NPWDataHeader const *)*addr)[-1].payloadSize))
		if (((NPWDataHeader const *)*addr)[-1].key == key)
			return kNPWErrOk;
	return kNPWErrDataNotFound;
}

NPWErr NPWDataGetNth(NPW *pdf, NPWConstText key, NPWInt nth, void **addr)
{
	NPWInt i;
	NPWErr err;
	
	*addr = NULL;
	for (i = 0; i <= nth; i++)
		Chk(NPWDataGetNext(pdf, key, (void **)addr));
	return kNPWErrOk;
}

void NPWDataDebug(NPW *pdf)
{
	void *headerAddr;
	NPWInt i;
	
	for (headerAddr = pdf->data, i = 0;
			headerAddr < (void *)((NPWByte *)pdf->data + pdf->dataSize);
			headerAddr = (void *)((NPWByte *)headerAddr + sizeof(NPWDataHeader)
					+ ((NPWDataHeader const *)headerAddr)->payloadSize), i++)
	{
		NPWWriteInt(pdf, i, -1);
		NPWWrite(pdf, " key=", -1);
		NPWWrite(pdf, ((NPWDataHeader const *)headerAddr)->key, -1);
		NPWWrite(pdf, " size=", -1);
		NPWWriteInt(pdf, ((NPWDataHeader const *)headerAddr)->payloadSize, -1);
		NPWWrite(pdf, " address=", -1);
		NPWWriteInt(pdf, (NPWInt)headerAddr + sizeof(NPWDataHeader), -1);
		NPWWrite(pdf, "\n", -1);
	}
}

NPWErr NPWBeginPage(NPW *pdf, NPWInt width, NPWInt height)
{
	NPWPage page, *pageAddr;
	NPWRef ref;
	NPWErr err;
	
	// write header if not done yet
	if (pdf->offset == 0)
		Chk(writeHeader(pdf));
	
	// end current page if not done
	if (pdf->inPage)
		Chk(NPWEndPage(pdf));
	
	// keep track of page
	page.width = width;
	page.height = height;
	Chk(NPWBeginObj(pdf, &page.contents));
	
	// begin contents
	Chk(NPWWrite(pdf, "<< /Length ", -1));
	Chk(reserveXRef(pdf, &ref));
	Chk(NPWWriteInt(pdf, ref, -1));
	Chk(NPWWrite(pdf, " 0 R >>\nstream\n", -1));
	page.streamOffset = pdf->offset;
	
	// store page
	Chk(NPWDataAlloc(pdf, kDataKeyPage, sizeof(NPWPage), (void **)&pageAddr));
	*pageAddr = page;
	
	pdf->inPage = TRUE;
	pdf->pageCount++;
	
	return kNPWErrOk;
}

NPWErr NPWEndPage(NPW *pdf)
{
	NPWPage *page;
	NPWInt length;
	NPWInt err;
	
	// check that a page has been begun
	if (!pdf->inPage)
		return kNPWErrUnexpectedEndPage;
	
	// end stream
	Chk(NPWDataGetNth(pdf, kDataKeyPage, pdf->pageCount - 1, (void **)&page));
	length = pdf->offset - page->streamOffset;
	Chk(NPWWrite(pdf, "endstream\n", -1));
	Chk(NPWEndObj(pdf));
	
	// write Length object
	Chk(beginReservedObj(pdf, NULL));
	Chk(NPWWriteInt(pdf, length, -1));
	Chk(NPWWrite(pdf, "\n", -1));
	Chk(NPWEndObj(pdf));
	return kNPWErrOk;
}

NPWBoolean NPWIsWritingPage(NPW *pdf)
{
	return pdf->inPage;
}

NPWErr NPWAddResourceCB(NPW *pdf, NPWActionCB cb, void *userData)
{
	NPWActionCBData *d;
	NPWErr err;
	
	Chk(NPWDataAlloc(pdf, kDataKeyRsrcCB, sizeof(NPWActionCBData), (void **)&d));
	d->cb = cb;
	d->userData = userData;
	return kNPWErrOk;
}

NPWErr NPWAddDelayedActionCB(NPW *pdf, NPWActionCB cb, void *userData)
{
	NPWActionCBData *d;
	NPWErr err;
	
	Chk(NPWDataAlloc(pdf, kDataKeyDelayedActionCB, sizeof(NPWActionCBData), (void **)&d));
	d->cb = cb;
	d->userData = userData;
	return kNPWErrOk;
}

NPWErr NPWTerminate(NPW *pdf)
{
	NPWOffset xrefTableOffset;
	NPWInt i;
	NPWActionCBData *delayedActionCBData;
	NPWRef pages, root;
	NPWXRef *xref;
	NPWErr err;
	
	// write header if not done yet
	if (pdf->offset == 0)
		ChkG(writeHeader(pdf));
	
	// end page if not done
	if (pdf->inPage)
		ChkG(NPWEndPage(pdf));
	
	// process delayed actions
	delayedActionCBData = NULL;
	for (;;)
	{
		if (NPWDataGetNext(pdf, kDataKeyDelayedActionCB, (void **)&delayedActionCBData) != kNPWErrOk)
			break;
		ChkG(delayedActionCBData->cb(pdf, delayedActionCBData->userData));
	}
	
	// write Pages object
	ChkG(writePages(pdf, &pages));
	
	// write Page objects
	for (i = 0; i < pdf->pageCount; i++)
		ChkG(writePage(pdf, i, pages, NULL));
	
	// write Catalog object
	ChkG(writeCatalog(pdf, pages, &root));
	
	// remember offset of cross-reference table
	xrefTableOffset = pdf->offset;
	
	// write cross-reference table
	NPWWrite(pdf, "xref\n0 ", -1);
	NPWWriteInt(pdf, pdf->xrefCount + 1, -1);	// including initial f entry
	NPWWrite(pdf, "\n", -1);
	NPWWrite(pdf, "0000000000 65535 f \n", -1);
	xref = NULL;
	for (i = 0; i < pdf->xrefCount; i++)
	{
		Chk(NPWDataGetNext(pdf, kDataKeyXRef, (void **)&xref));
		NPWWriteInt(pdf, xref->offset, 10);
		NPWWrite(pdf, " ", -1);
		NPWWriteInt(pdf, xref->generation, 5);
		NPWWrite(pdf, " n \n", -1);
	}
	
	// write trailer dictionary
	NPWWrite(pdf, "trailer\n<<\n", -1);
	NPWWrite(pdf, "/Size ", -1);
	NPWWriteInt(pdf, pdf->xrefCount, -1);
	NPWWrite(pdf, "\n", -1);
	NPWWrite(pdf, "/Root ", -1);
	NPWWriteInt(pdf, root, -1);
	NPWWrite(pdf, " 0 R\n", -1);
	if (pdf->info != kNPWNoRef)
	{
		ChkG(NPWWrite(pdf, "/Info ", -1));
		ChkG(NPWWriteInt(pdf, pdf->info, -1));
		ChkG(NPWWrite(pdf, " 0 R\n", -1));
	}
	NPWWrite(pdf, ">>\n", -1);
	
	// write xref offset
	NPWWrite(pdf, "startxref\n", -1);
	NPWWriteInt(pdf, xrefTableOffset, -1);
	NPWWrite(pdf, "\n", -1);
	
	// write last line
	NPWWrite(pdf, "%%EOF\n", -1);
	
error:
	// free data
	if (pdf->alloc && pdf->data)
		(void)pdf->alloc((void **)&pdf->data, -1, pdf->allocUserData);
	
	return err;
}
