/**
 *	@file NPWBase.h
 *	@brief NPW (base functions to write PDF files)
 *	@author Yves Piguet.
 *	@copyright 2008-2014, Yves Piguet
 */

/** @mainpage
 *
 *	Nyctergatis PDF Writer (NPW) - C library to write simple PDF files.
 *
 *	@author Yves Piguet
 *	@see http://nyctergatis.com (home of NPW project)
 *
 *	The main goal of NPW is simplicity, to the cost of completeness.
 *	PDF files are more difficult to write than Postscript files, because
 *	they require cross-reference tables and the computation of file
 *	offsets.
 *
 *	@section Features Features
 *
 *	NPW is made of a base set of functions and data structures,
 *	NPWBase, which can be used as a wrapper around the graphics and
 *	text commands which define page contents. NPWGraphics provides
 *	support for most of the graphical and textual PDF commands.
 *	NPWText adds support for text layout based on font metrics,
 *	including wordwrap and text flow across multiple rectangles and
 *	pages, with optional support for markup to have multiple character
 *	styles and paragraph formats; NPWTextMarkup implements such a
 *	markup. NPWInfo adds support for information metadata such as
 *	author and creation date.
*/

/**	@page NPWusage NPW Usage
 *
 *	NPW requires the definition of two callback functions, one for
 *	memory allocation/reallocation/deallocation, and one for output.
 *	Callbacks based on malloc/realloc/free and fwrite are defined in
 *	NPWStdAlloc.c and NPWStdOutput.c, respectively.
 *
 *	Here is how NPW's main data structure, of type NPW, can be initialized
 *	to write a PDF file as "doc.pdf":
 *
 *	@code
#include "NPWBase.h"
#include "NPWStdAlloc.h"
#include "NPWStdOutput.h"
#include <stdio.h>
 *
NPW npw;
FILE *fp;
fp = fopen("doc.pdf", "wb");
NPWInit(&npw);
NPWSetAllocCallback(&npw, NPWAllocStdlib, NULL);
NPWSetWriteCallback(&npw, NPWWriteStdOutput, (void *)fp);
 *	@endcode
 *
 *	Each page is added with a pair of calls to NPWBeginPage and
 *	NPWEndPage, with commands which produce displayed material
 *	inbetween. Each page can have a different size, given (like
 *	all other sizes and positions) in points (1/72 inch).
 *
 *	@code
NPWBeginPage(&npw, kNPWPaperWidthA4, kNPWPaperHeightA4);
...
NPWEndPage(&npw);
 *	@endcode
 *
 *	To add graphical commands, you can either do it by hand with calls
 *	to NPWWrite, NPWWriteInt, NPWWriteFloat or NPWWriteString, or
 *	rely on commands defined in NPWGraphics.h:
 *
 *	@code
NPWWrite(&npw, "10 10 m 100 100 l S\n", -1);
 *	@endcode
 *
 *	is equivalent to
 *
 *	@code
#include "NPWGraphics.h"
NPWGraphicsMove(&npw, 10, 10);
NPWGraphicsLine(&npw, 100, 100);
NPWGraphicsStroke(&npw);
 *	@endcode
 *
 *	To add meta-informations such as the author or the creation date,
 *	add calls to NPWInfoAddEntry anywhere:
 *
 *	@code
#include "NPWInfo.h"
NPWInfoAddEntry(&npw, kNPWInfoKeyTitle, "The Title");
NPWInfoAddEntry(&npw, kNPWInfoKeyAuthor, "The Author");
 *	@endcode
 *
 *	Finally, to terminate the PDF and deallocate data structures allocated
 *	with previous calls, the NPWTerminate function is executed:
 *
 *	@code
NPWTerminate(&npw);
fclose(fp);
 *	@endcode
 */

/**	@page NPWlicense NPW License
 *
 *	Copyright (c) 2007-2014, Yves Piguet
 *	All rights reserved.
 *
 *	Redistribution and use in source and binary forms, with or without
 *	modification, are permitted provided that the following conditions
 *	are met:
 *
Redistributions of source code must retain the above copyright notice,
this list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

Neither the name of Yves Piguet nor the names of its contributors may be
used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	@section Remark Remarks
 *
 *	Sorry for shouting; we keep the new BSD license disclaimer verbatim.
 *
 *	The license above is known as the "new BSD license", i.e. without
 *	advertising clause. For a discussion on different open-source and
 *	"free" ("copyleft") licenses, please visit  http://www.opensource.org
 *	and http://www.gnu.org.
 */

#ifndef __NPWBase__
#define __NPWBase__

#ifdef __cplusplus
extern "C" {
#endif

/// Character (8-bit; processing possible with compatible charsets like UTF-8 or Shift-JIS)
typedef char NPWChar;

/// Text
typedef NPWChar *NPWText;

/// Constant text
typedef NPWChar const *NPWConstText;

/// Integer
typedef int NPWInt;

/// Object reference
typedef NPWInt NPWRef;

/// Offset in PDF file
typedef NPWInt NPWOffset;

/// Unsigned 8-bit integer
typedef unsigned char NPWByte;

/// Unsigned 32-bit integer
typedef unsigned long NPWUInt32;

/// Signed 16-bit integer
typedef short PDFInt16;

// Floating-point number
#if !defined(NPWNoFloat)
/// Floating-point number
typedef double NPWFloat;
#else
/// Real number represented by an integer on platforms which lack floating-point
typedef NPWInt NPWFloat;
#endif

/// Boolean
typedef int NPWBoolean;

#if !defined(NULL)
/// null pointer
#	define NULL 0
#endif

#if !defined(FALSE)
#	define FALSE 0	///< false boolean value
#	define TRUE 1	///< true boolean value
#endif

/// Special references
enum
{
	kNPWNoRef = -1	///< Special value for nonexistent reference
};

/// Error codes
typedef enum
{
	kNPWErrOk = 0,	///< no error
	kNPWErrNotEnoughMemory,	///< not enough memory
	kNPWErrUnexpectedEndPage,	///< NPWEndPage without matching NPWBeginPage
	kNPWErrUnexpectedCommand,	///< graphical command not between NPWBeginPage and NPWEndPage
	kNPWErrNestedObject,	///< try to start a new object without having ended previous one
	kNPWErrUnbalancedObj,	///< endobj without obj
	kNPWUnknownFont,	///< unknown font
	kNPWNumTooLarge,	///< floating-point number too large
	kNPWErrIncomplete,	///< incomplete operation
	kNPWErrSuspended,	///< operation suspended by callback to permit a lower-level action
	kNPWErrDataNotFound,	///< NPWDataGetNext cannot find more data
	kNPWErrFileIO,	///< file I/O error
	
	kNPWErrInternal	///< internal error (should never occur)
} NPWErr;

/// Paper size
enum
{
	// ISO 216
	kNPWPaperHeightA3 = 1191,	///< portrait A3 height (420 mm)
	kNPWPaperWidthA3 = 842,	///< portrait A3 width (297 mm)
	kNPWPaperHeightA4 = 842,	///< portrait A4 height (297 mm)
	kNPWPaperWidthA4 = 595,	///< portrait A4 width (210 mm)
	kNPWPaperHeightA5 = 595,	///< portrait A5 height (210 mm)
	kNPWPaperWidthA5 = 420,	///< portrait A5 width (148 mm)
	kNPWPaperHeightA6 = 420,	///< portrait A6 height (148 mm)
	kNPWPaperWidthA6 = 298,	///< portrait A6 width (105 mm)
	
	kNPWPaperHeightB3 = 1417,	///< portrait B3 height (500 mm)
	kNPWPaperWidthB3 = 1001,	///< portrait B3 width (353 mm)
	kNPWPaperHeightB4 = 1001,	///< portrait B4 height (353 mm)
	kNPWPaperWidthB4 = 709,	///< portrait B4 width (250 mm)
	kNPWPaperHeightB5 = 709,	///< portrait B5 height (250 mm)
	kNPWPaperWidthB5 = 499,	///< portrait B5 width (176 mm)
	kNPWPaperHeightB6 = 499,	///< portrait B6 height (176 mm)
	kNPWPaperWidthB6 = 354,	///< portrait B6 width (125 mm)
	
	kNPWPaperHeightC3 = 1298,	///< portrait C3 height (458 mm)
	kNPWPaperWidthC3 = 918,	///< portrait C3 width (324 mm)
	kNPWPaperHeightC4 = 918,	///< portrait C4 height (324 mm)
	kNPWPaperWidthC4 = 649,	///< portrait C4 width (229 mm)
	kNPWPaperHeightC5 = 649,	///< portrait C5 height (229 mm)
	kNPWPaperWidthC5 = 459,	///< portrait C5 width (162 mm)
	kNPWPaperHeightC6 = 459,	///< portrait C6 height (162 mm)
	kNPWPaperWidthC6 = 323,	///< portrait C6 width (114 mm)
	
	kNPWPaperHeightDL = 312,	///< DL envelope height(110 mm)
	kNPWPaperWidthDL = 624,	///< DL envelope width (220 mm)
	
	// ISO 7810
	kNPWPaperHeightID1 = 153,	///< ID-1 (banking cards) height (53.98 mm)
	kNPWPaperWidthID1 = 243,	///< ID-1 (banking card) width (85.6 mm)
	kNPWPaperHeightID2 = 210,	///< ID-2 (A7) height (74 mm)
	kNPWPaperWidthID2 = 298,	///< ID-2 (A7) width (105 mm)
	kNPWPaperHeightID3 = 249,	///< ID-3 (B7, passport) height (88 mm)
	kNPWPaperWidthID3 = 354,	///< ID-3 (B7, passport) width (125 mm)
	
	// US
	kNPWPaperHeightLetter = 792,	///< Letter height (279 mm)
	kNPWPaperWidthLetter = 612,	///< Letter height (216 mm)
	kNPWPaperHeightLegal = 1008,	///< Legal height (356 mm)
	kNPWPaperWidthLegal = 612,	///< Legal height (216 mm)
	kNPWPaperHeightExecutive = 756,	///< Executive height (266.7 mm)
	kNPWPaperWidthExecutive = 522,	///< Executive height (184.2 mm)
	kNPWPaperHeightLedger = 1224,	///< Ledger/Tabloid height (431.8 mm)
	kNPWPaperWidthLedger = 792	///< Ledger/Tabloid height (279.4 mm)
};

// Unit conversion factors
#define kNPWUnitMM 2.835	///< conversion factor from mm to pt
#define kNPWUnitCM 28.35	///< conversion factor from cm to pt
#define kNPWUnitInch 72	///< conversion factor from inch to pt

struct NPW_struct;	// forward reference

/**	Callback to write data to PDF file.
	@param[in] str address of string
	@param[in] length length of string in bytes
	@param userData application-specific data
	@return kNPWErrOk for success, else error code
*/
typedef NPWErr (*NPWWriteCB)(NPWConstText str, NPWInt length, void *userData);

/**	Callback to allocate, resize or free block of memory.
	@param[in,out] p pointer to memory (NULL to allocate new block)
	@param[in] size size of block (negative to free it)
	@param userData application-specific data
	@return kNPWErrOk for success, else error code
*/
typedef NPWErr (*NPWAllocCB)(void **p, NPWInt size, void *userData);

/**	Callback to perform some action.
	@param pdf reference to NPW structure
	@param userData application-specific data
	@return kNPWErrOk for success, else error code
*/
typedef NPWErr (*NPWActionCB)(struct NPW_struct *pdf, void *userData);

/// Cross-reference description
typedef struct
{
	NPWOffset offset;	///< offset in PDF file
	NPWInt generation;	///< generation number
} NPWXRef;

/// Page description
typedef struct
{
	NPWInt width;	///< page width
	NPWInt height;	///< page height
	NPWRef contents;	///< reference to page contents object
	NPWInt streamOffset;	///< offset of beginning of contents stream
} NPWPage;

/// Filter flags (encoding: first rl, then hex or 85)
enum
{
	kNPWFilterNone = 0,	///< no filter
	kNPWFilterASCIIHex = 1,	///< ASCII Hexadecimal filter
	kNPWFilterASCII85 = 2,	///< ASCII base-85 filter
	kNPWFilterRunLength = 4	///< run-length compression filter
};

#define NPWFilterNameASCIIHex "/AHx"	///< hexadecimal filter name
#define NPWFilterNameASCII85 "/A85"	///< base-85 filter name
#define NPWFilterNameRunLength "/RL"	///< run-length filter name
#define NPWFilterNameDCT "/DCT"	///< DCT filter name (JPEG)

///	Callback and user data for performing an action
typedef struct
{
	NPWActionCB cb;	///< callback
	void *userData;	///< user data
} NPWActionCBData;

/// NPW main structure
typedef struct NPW_struct
{
	NPWInt width;	///< default page width
	NPWInt height;	///< default page height
	NPWInt fonts;	///< ored kNPWFontCourier etc.
	
	NPWAllocCB alloc;	///< memory allocation callback
	void *allocUserData;	///< memory allocation callback user data
	
	NPWWriteCB write;	///< PDF output callback
	void *writeUserData;	///< PDF output callback user data
	
	NPWOffset offset;	///< offset of next byte written to PDF file
	
	NPWBoolean inPage;	///< TRUE between NPWBeginPage and NPWEndPage
	NPWBoolean writingObject;	///< TRUE while writing an object, to prevent nesting
	
	void *data;	///< data allocated by NPWDataAlloc
	NPWInt dataSize;	///< current size of used data
	NPWInt dataAllocatedSize;	///< current size allocated to data
	
	NPWInt xrefCount;	///< current number of xref in data
	NPWInt pageCount;	///< current number of pages in data
	
	NPWRef info;	///< reference of document information dictionary, or kNPWNoRef
	
	NPWInt filter;	///< kNPWFilterNone etc.
	NPWByte asciiBuf[4];	///< bytes not yet encoded with ASCII85Decode filter
	NPWInt asciiBufCount;	///< number of bytes in asciiBuf
	NPWInt asciiFilterLineLength;	///< number of characters in line of ASCII-filtered stream
	
} NPW;

/// Standard font flags
enum
{
	kNPWFontCourier = 1,	///< Courier
	kNPWFontCourierOblique = 2,	///< Courier oblique
	kNPWFontCourierBold = 4,	///< Courier bold
	kNPWFontCourierBoldOblique = 010,	///< Courier bold oblique
	kNPWFontHelvetica = 020,	///< Helvetica
	kNPWFontHelveticaOblique = 040,	///< Helvetica oblique
	kNPWFontHelveticaBold = 0100,	///< Helvetica bold
	kNPWFontHelveticaBoldOblique = 0200,	///< Helvetica bold oblique
	kNPWFontTimesRoman = 0400,	///< Times roman
	kNPWFontTimesItalic = 01000,	///< Times italic
	kNPWFontTimesBold = 02000,	///< Times bold
	kNPWFontTimesBoldItalic = 04000,	///< Times bold italic
	kNPWFontSymbol = 010000,	///< Symbol
	kNPWFontZapfDingbats = 020000,	///< Zapf Dingbats
	kNPWFontNone = 0,	///< no font
	kNPWFontAll = 037777	///< all fonts
};

#define NPWFontNameCourier "/FC"	///< Courier font name
#define NPWFontNameCourierOblique "/FCo"	///< Courier Oblique font name
#define NPWFontNameCourierBold "/FCb"	///< Courier Bold font name
#define NPWFontNameCourierBoldOblique "/FCbo"	///< Courier Bold Oblique font name
#define NPWFontNameHelvetica "/FH"	///< Helvetica font name
#define NPWFontNameHelveticaOblique "/FHo"	///< Helvetica Oblique font name
#define NPWFontNameHelveticaBold "/FHb"	///< Helvetica Bold font name
#define NPWFontNameHelveticaBoldOblique "/FHbo"	///< Helvetica Bold Oblique font name
#define NPWFontNameTimesRoman "/FT"	///< Times Roman font name
#define NPWFontNameTimesItalic "/FTi"	///< Time Italic font name
#define NPWFontNameTimesBold "/FTb"	///< Times Bold font name
#define NPWFontNameTimesBoldItalic "/FTbi"	///< Times Bold Italic font name
#define NPWFontNameSymbol "/FS"	///< Symbol font name
#define NPWFontNameZapfDingbats "/FZ"	///< Zapf Dingbats font name

/**	Initialize PDF.
	@param[out] pdf reference to NPW structure
*/
void NPWInit(NPW *pdf);

/** Set memory allocation callback.
	@param pdf reference to NPW structure
	@param[in] alloc memory allocation callback
	@param[in] userData callback user data
*/
void NPWSetAllocCallback(NPW *pdf,
		NPWAllocCB alloc, void *userData);

/** Set write callback.
	@param pdf reference to NPW structure
	@param[in] write allocation callback
	@param[in] userData callback user data
*/
void NPWSetWriteCallback(NPW *pdf,
		NPWWriteCB write, void *userData);

/** Specify which standard fonts should be made available.
	@param pdf reference to NPW structure
	@param[in] fonts font flags (kNPWFontAll for all)
*/
void NPWSetFonts(NPW *pdf, NPWInt fonts);

/**	Terminate PDF and flush output.
	@param pdf reference to NPW structure
	@return kNPWErrOk for success, else error code
*/
NPWErr NPWTerminate(NPW *pdf);

/**	Allocate a block of data which can be retrieved by key and index.
	@param pdf reference to NPW structure
	@param[in] key key (whose address is stored)
	@param[in] size size in bytes
	@param[out] addr address of allocated data
	@return kNPWErrOk for success, else error code
*/
NPWErr NPWDataAlloc(NPW *pdf, NPWConstText key, NPWInt size, void **addr);

/**	Count number of data of a given type.
	@param pdf reference to NPW structure
	@param[in] key key (whose address is used)
	@return number of data
*/
NPWInt NPWDataCount(NPW *pdf, NPWConstText key);

/**	Get next data of a given type.
	@param pdf reference to NPW structure
	@param[in] key key (whose address is used)
	@param[in,out] addr data address (NULL on input to get first one)
	@return kNPWErrOk for success, kNPWErrDataNotFound if not found
*/
NPWErr NPWDataGetNext(NPW *pdf, NPWConstText key, void **addr);

/**	Get nth data of a given type.
	@param pdf reference to NPW structure
	@param[in] key key (whose address is used)
	@param[in] nth index of the data to retrieve
	@param[out] addr data address
	@return kNPWErrOk for success, kNPWErrDataNotFound if not found
*/
NPWErr NPWDataGetNth(NPW *pdf, NPWConstText key, NPWInt nth, void **addr);

/**	Display (add to output) a list of data of all types.
	@param pdf reference to NPW structure
*/
void NPWDataDebug(NPW *pdf);

/**	Begin a new page (end the current page first if NPWEndPage hasn't been called).
	@param pdf reference to NPW structure
	@param[in] width page width
	@param[in] height page height
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWBeginPage(NPW *pdf, NPWInt width, NPWInt height);

/**	End the current page.
	@param pdf reference to NPW structure
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWEndPage(NPW *pdf);

/**	Check if is currently writing a page
	@param pdf reference to NPW structure
	@return TRUE if between NPWBeginPage and NPWEndPage, else FALSE
*/
NPWBoolean NPWIsWritingPage(NPW *pdf);

/**	Add a callback which will be called at appropriate time to add
	resources to the Resources dictionary.
	@param pdf reference to NPW structure
	@param[in] cb callback
	@param[in] userData callback user data
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWAddResourceCB(NPW *pdf, NPWActionCB cb, void *userData);

/**	Add a callback which will be called at the end, after writing the pages.
	@param pdf reference to NPW structure
	@param[in] cb callback
	@param[in] userData callback user data
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWAddDelayedActionCB(NPW *pdf, NPWActionCB cb, void *userData);

/**	Write a string.
	@param pdf reference to NPW structure
	@param[in] str address of first character
	@param[in] length number of bytes, or -1 if str is null-terminated
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWWrite(NPW *pdf, NPWConstText str, NPWInt length);

/**	Write an integer.
	@param pdf reference to NPW structure
	@param[in] i number
	@param[in] pad number of digits (0-padded), or -1
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWWriteInt(NPW *pdf, NPWInt i, NPWInt pad);

#if !defined(NPWNoFloat)
/**	Write an floating-point value.
	@param pdf reference to NPW structure
	@param[in] x number
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWWriteFloat(NPW *pdf, NPWFloat x);
#endif

/**	Write a string value (should be ISO Latin 1 to match encoding set for standard fonts).
	@param pdf reference to NPW structure
	@param[in] str string
	@param[in] length string length in bytes, or -1 for null-terminated string
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWWriteString(NPW *pdf, NPWConstText str, NPWInt length);

/**	Begin writing an object.
	@param pdf reference to NPW structure
	@param[out] ref object reference number
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWBeginObj(NPW *pdf, NPWRef *ref);

/**	End writing an object.
	@param pdf reference to NPW structure
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWEndObj(NPW *pdf);

/**	Begin writing filtered data to PDF.
	@param pdf reference to NPW structure
	@param[in] filter (kNPWFilterNone for none)
*/
void NPWFilterBegin(NPW *pdf, NPWInt filter);

/**	Write filtered data to PDF.
	@param pdf reference to NPW structure
	@param[in] b address of unfiltered data
	@param[in] n number of unfiltered bytes
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWFilterWrite(NPW *pdf, NPWByte const *b, NPWInt n);

/**	End writing filtered data to PDF.
	@param pdf reference to NPW structure
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWFilterEnd(NPW *pdf);

#ifdef __cplusplus
}
#endif

#endif
