/**
 *	@file NPWGraphics.h
 *	@brief Helper functions for graphics
 *	@author Yves Piguet.
 *	@copyright 2008-2014, Yves Piguet
 */

#ifndef __NPWGraphics__
#define __NPWGraphics__

#ifdef __cplusplus
extern "C" {
#endif

#include "NPWBase.h"

/**	Write arguments and a command.
	@param pdf reference to NPW structure
	@param[in] cmd PDF command
	@param[in] fmt argument types (f=float, i=int, S=raw string+len, s=PDF string+len)
	@return error code (kNPWErrOk for success)
*/
NPWErr NPWGraphicsCmd(NPW *pdf, NPWConstText cmd, NPWConstText fmt, ...);

/// Path "move" command
#define NPWGraphicsMove(pdf, x, y) \
	NPWGraphicsCmd(pdf, "m", "ff", (NPWFloat)(x), (NPWFloat)(y))
/// Path "line" command
#define NPWGraphicsLine(pdf, x, y) \
	NPWGraphicsCmd(pdf, "l", "ff", (NPWFloat)(x), (NPWFloat)(y))
/// Path "bezier" command
#define NPWGraphicsBezier(pdf, x1, y1, x2, y2, x3, y3) \
	NPWGraphicsCmd(pdf, "c", "ffffff", \
		(NPWFloat)(x1), (NPWFloat)(y1), (NPWFloat)(x2), (NPWFloat)(y2), (NPWFloat)(x3), (NPWFloat)(y3))
/// Close path
#define NPWGraphicsClosePath(pdf) \
	NPWGraphicsCmd(pdf, "h", "")
/// Create a rectangular path
#define NPWGraphicsRectangle(pdf, x, y, w, h) \
	NPWGraphicsCmd(pdf, "re", "ffff", (NPWFloat)(x), (NPWFloat)(y), (NPWFloat)(w), (NPWFloat)(h))

/// Create an elliptic path
#define NPWGraphicsEllipse(pdf, x0, y0, rx, ry) \
	do { \
		NPWGraphicsMove(pdf, (x0) - (rx), y0); \
		NPWGraphicsBezier(pdf, \
				(x0) - (rx), (y0) + (ry) * 55 / 100, \
				(x0) - (rx) * 55 / 100, (y0) + (ry), \
				x0, (y0) + (ry)); \
		NPWGraphicsBezier(pdf, \
				(x0) + (rx) * 55 / 100, (y0) + (ry), \
				(x0) + (rx), (y0) + (ry) * 55 / 100, \
				(x0) + (rx), y0); \
		NPWGraphicsBezier(pdf, \
				(x0) + (rx), (y0) - (ry) * 55 / 100, \
				(x0) + (rx) * 55 / 100, (y0) - (ry), \
				x0, (y0) - (ry)); \
		NPWGraphicsBezier(pdf, \
				(x0) - (rx) * 55 / 100, (y0) - (ry), \
				(x0) - (rx), (y0) - (ry) * 55 / 100, \
				(x0) - (rx), y0); \
		NPWGraphicsClosePath(pdf); \
	} while (0)

/// Stroke path
#define NPWGraphicsStroke(pdf) \
	NPWGraphicsCmd(pdf, "S", "")
/// Close and stroke path
#define NPWGraphicsCloseStroke(pdf) \
	NPWGraphicsCmd(pdf, "s", "")
/// Fill path using nonzero winding number rule
#define NPWGraphicsFill(pdf) \
	NPWGraphicsCmd(pdf, "f", "")
/// Fill path using even-odd rule
#define NPWGraphicsFillEO(pdf) \
	NPWGraphicsCmd(pdf, "f*", "")
/// Fill path using nonzero winding number rule, then stroke it
#define NPWGraphicsFillStroke(pdf) \
	NPWGraphicsCmd(pdf, "B", "")
/// Fill path using even-odd rule, then stroke it
#define NPWGraphicsFillEOStroke(pdf) \
	NPWGraphicsCmd(pdf, "B*", "")
/// Close path, fill it using nonzero winding number rule, then stroke it
#define NPWGraphicsCloseFillStroke(pdf) \
	NPWGraphicsCmd(pdf, "b", "")
/// Close path, fill it using even-odd rule, then stroke it
#define NPWGraphicsCloseFillEOStroke(pdf) \
	NPWGraphicsCmd(pdf, "b*", "")
/// Discard path
#define NPWGraphicsEndPath(pdf) \
	NPWGraphicsCmd(pdf, "n", "")

/// Use path to define the clipping region with nonzero winding number rule
#define NPWGraphicsClip(pdf) \
	NPWGraphicsCmd(pdf, "W", "")
/// Use path to define the clipping region with even-odd rule
#define NPWGraphicsClipEO(pdf) \
	NPWGraphicsCmd(pdf, "W*", "")

/// Define the gray shade for stroking
#define NPWGraphicsGrayStroke(pdf, g) \
	NPWGraphicsCmd(pdf, "G", "f", (NPWFloat)(g))
/// Define the gray shade for filling
#define NPWGraphicsGray(pdf, g) \
	NPWGraphicsCmd(pdf, "g", "f", (NPWFloat)g)
/// Define the RGB color for stroking
#define NPWGraphicsRGBStroke(pdf, r, g, b) \
	NPWGraphicsCmd(pdf, "RG", "fff", (NPWFloat)(r), (NPWFloat)(g), (NPWFloat)(b))
/// Define the RGB color for filling
#define NPWGraphicsRGB(pdf, r, g, b) \
	NPWGraphicsCmd(pdf, "rg", "fff", (NPWFloat)(r), (NPWFloat)(g), (NPWFloat)(b))
/// Define the CMYK color for stroking
#define NPWGraphicsCMYKStroke(pdf, c, m, y, k) \
	NPWGraphicsCmd(pdf, "K", "ffff", (NPWFloat)(c), (NPWFloat)(m), (NPWFloat)(y), (NPWFloat)(k))
/// Define the CMYK color for filling
#define NPWGraphicsCMYK(pdf, c, m, y, k) \
	NPWGraphicsCmd(pdf, "k", "ffff", (NPWFloat)(c), (NPWFloat)(m), (NPWFloat)(y), (NPWFloat)(k))

/// Line cap type
enum
{
	kNPWLineCapButt = 0,	///< butt cap
	kNPWLineCapRound,	///< round cap
	kNPWLineCapSquare	///< square cap
};

/// Line join type
enum
{
	kNPWLineJoinMitter = 0,	///< mitter join
	kNPWLineJoinRound,	///< round join
	kNPWLineJoinBevel	///< bevel join
};

/// Save graphics state
#define NPWGraphicsSave(pdf) \
	NPWGraphicsCmd(pdf, "q", "")
/// Restore graphics state
#define NPWGraphicsRestore(pdf) \
	NPWGraphicsCmd(pdf, "Q", "")
/// Concatenate a matrix to the current coordinate transform matrix
#define NPWGraphicsConcatCTM(pdf, a, b, c, d, e, f) \
	NPWGraphicsCmd(pdf, "cm", "ffffff", \
		(NPWFloat)(a), (NPWFloat)(b), (NPWFloat)(c), (NPWFloat)(d), (NPWFloat)(e), (NPWFloat)(f))
/// Concatenate a translation matrix to the current coordinate transform matrix
#define NPWGraphicsTranslate(pdf, dx, dy) \
	NPWGraphicsConcatCTM(pdf, 1, 0, 0, 1, dx, dy)
/// Concatenate a scaling matrix to the current coordinate transform matrix
#define NPWGraphicsScale(pdf, sx, sy) \
	NPWGraphicsConcatCTM(pdf, sx, 0, 0, sy, 0, 0)
/// Set line width for stroking
#define NPWGraphicsSetLineWidth(pdf, w) \
	NPWGraphicsCmd(pdf, "w", "f", (NPWFloat)(w))
/// Set line cap
#define NPWGraphicsSetLineCap(pdf, c) \
	NPWGraphicsCmd(pdf, "J", "i", c)
/// Set line join
#define NPWGraphicsSetLineJoin(pdf, j) \
	NPWGraphicsCmd(pdf, "j", "i", j)
/// Set miter limit
#define NPWGraphicsSetMiterLimit(pdf, l) \
	NPWGraphicsCmd(pdf, "M", "f", (NPWFloat)(l))

/// Begin an inline raster image (must be followed by the raster data
/// written with NPWFilterBegin, NPWFilterWrite and NPWFilterEnd)
#define NPWInlineImageBegin(pdf, w, h, cs, bpc, f) \
	do { \
		NPWWrite(pdf, "BI\n/W ", -1); \
		NPWWriteInt(pdf, w, -1); \
		NPWWrite(pdf, "\n/H ", -1); \
		NPWWriteInt(pdf, h, -1); \
		NPWWrite(pdf, "\n/CS ", -1); \
		NPWWrite(pdf, cs, -1); \
		NPWWrite(pdf, "\n/BPC ", -1); \
		NPWWriteInt(pdf, bpc, -1); \
		NPWWrite(pdf, "\n/F ", -1); \
		NPWWrite(pdf, f, -1); \
		NPWWrite(pdf, "\nID ", -1); \
	} while (0)
/// End an inline image
#define NPWInlineImageEnd(pdf) \
	NPWWrite(pdf, "EI\n", -1);

/// Text rendering modes
enum
{
	kNPWTextRenderingModeFill = 0,	///< fill text (normal)
	kNPWTextRenderingModeStroke,	///< stroke text
	kNPWTextRenderingModeFillStroke,	///< fill, then stroke text
	kNPWTextRenderingModeInvisible,	///< do nothing (invisible)
	kNPWTextRenderingModeFillClip,	///< fill text and add to path for clipping
	kNPWTextRenderingModeStrokeClip,	///< stroke text and add to path for clipping
	kNPWTextRenderingModeFillStrokeClip,	///< fill, then stroke text and add to path for clipping
	kNPWTextRenderingModeClip	///< add text to path for clipping
};

/// Set additional space between each character pair (positive to reduce space)
#define NPWGraphicsSetCharSpace(pdf, s) \
	NPWGraphicsCmd(pdf, "Tc", "f", (NPWFloat)(s))
/// Set additional space between each word
#define NPWGraphicsSetWordSpace(pdf, s) \
	NPWGraphicsCmd(pdf, "Tw", "f", (NPWFloat)(s))
/// Set horizontal scaling
#define NPWGraphicsSetHorizontalScaling(pdf, s) \
	NPWGraphicsCmd(pdf, "Tz", "f", (NPWFloat)(s))
/// Set text leading (distance between lines, negative to go downward)
#define NPWGraphicsSetTextLeading(pdf, leading) \
	NPWGraphicsCmd(pdf, "TL", "f", (NPWFloat)(leading))
/// Set text font (NPWFontNameCourier etc.) and size
#define NPWGraphicsSetTextFont(pdf, font, size) \
	NPWGraphicsCmd(pdf, "Tf", "Sf", font, -1, (NPWFloat)(size))
/// Set text rendering mode (default = kNPWTextRenderingModeFill)
#define NPWGraphicsSetTextRenderingMode(pdf, render) \
	NPWGraphicsCmd(pdf, "Tr", "f", (NPWFloat)(render))
/// Set text rise (positive for superscript)
#define NPWGraphicsSetTextRise(pdf, rise) \
	NPWGraphicsCmd(pdf, "Ts", "f", (NPWFloat)(rise))

/// Begin writing text
#define NPWGraphicsBeginText(pdf) \
	NPWGraphicsCmd(pdf, "BT", "")
/// End writing text
#define NPWGraphicsEndText(pdf) \
	NPWGraphicsCmd(pdf, "ET", "")
/// Move position of next text drawing
#define NPWGraphicsMoveText(pdf, x, y) \
	NPWGraphicsCmd(pdf, "Td", "ff", (NPWFloat)(x), (NPWFloat)(y))
/// Move to beginning of next line
#define NPWGraphicsNL(pdf) \
	NPWGraphicsCmd(pdf, "T*", "")
/// Show text
#define NPWGraphicsShowText(pdf, str, len) \
	NPWGraphicsCmd(pdf, "Tj", "s", str, len)
/// Move to beginning of next line, then show text
#define NPWGraphicsNLShowText(pdf, str, len) \
	NPWGraphicsCmd(pdf, "'", "s", str, len)
/// Move to beginning of next line, set space between words and characters, then show text
#define NPWGraphicsNLShowTextSp(pdf, aw, ac, str, len) \
	NPWGraphicsCmd(pdf, "\"", "ffs", (NPWFloat)(aw), (NPWFloat)(ac), str, len)
/// Adjust horizontal displacement by moving to the right, in 1/1000 font size
#define NPWGraphicsHorSkip(pdf, dx) \
	NPWGraphicsCmd(pdf, "TJ", "[f]", -(NPWFloat)dx)

#ifdef __cplusplus
}
#endif

#endif
