/**
 *	@file NPWMetrics.h
 *	@brief Font metrics
 *	@author Yves Piguet.
 *	@copyright 2008-2013, Yves Piguet
 */

#ifndef __NPWMetrics__
#define __NPWMetrics__

#ifdef __cplusplus
extern "C" {
#endif

#include "NPWBase.h"

/// Font metrics
typedef struct
{
	NPWFloat size;	///< font size
	void const *metrics;	///< opaque pointer to font metrics
} NPWFont;

/// Vertical character alignment
typedef enum
{
	kNPWCharVertAlignmentNormal = 0,	///< normal text
	kNPWCharVertAlignmentSubscript,	///< subscript
	kNPWCharVertAlignmentSuperscript	///< superscript
} NPWCharVertAlignment;

/// RGB color
typedef struct
{
	NPWByte r;	///< red component
	NPWByte g;	///< green component
	NPWByte b;	///< blue component
} NPWRGBColor;

/// Character style
typedef struct
{
	NPWConstText fontBaseName;	///< font base name (Times, Helvetica, Courier, Symbol or ZapfDingbats)
	NPWFloat size;	///< font size
	NPWBoolean isBold;	///< TRUE for bold
	NPWBoolean isItalic;	///< TRUE for italic
	NPWBoolean isUnderlined;	///< TRUE for underlined
	NPWBoolean isHidden;	///< TRUE for hidden (not displayed, except underline)
	NPWCharVertAlignment verticalAlignment;	///< character vertical alignment
	NPWRGBColor color;	///< character color
} NPWCharStyle;

/**	Initialize a font by name and size.
	@param[out] font font
	@param[in] name font dictionary name (NPWFontNameCourier etc.)
	@param[in] size font size in points
	@return kNPWErrOk for success, else error code
*/
NPWErr NPWTextFont(NPWFont *font, NPWConstText name, NPWFloat size);

/**	Calculate size of text.
	@param[in] font font
	@param[in] str text
	@param[in] len length of text, or -1 if str is null-terminated
	@param[out] width text width if not NULL
	@param[out] height text height if not NULL
*/
void NPWTextSize(NPWFont const *font,
		NPWConstText str, NPWInt len,
		NPWFloat *width, NPWFloat *height);

/**	Find a font by name and style.
	@param[in] charStyle font and character style
	@param[out] font AFM metrics
	@return font reference to be used in PDF
*/
NPWConstText NPWFindFont(NPWCharStyle const *charStyle, NPWFont *font);

/**	Get font metrics.
	@param[in] font AFM metrics
	@param[out] ascender ascender (not retrieved if NULL)
	@param[out] descender descender (not retrieved if NULL)
	@param[out] xHeight x-height (not retrieved if NULL)
*/
void NPWFontMetrics(NPWFont const *font,
		NPWFloat *ascender,
		NPWFloat *descender,
		NPWFloat *xHeight);

#ifdef __cplusplus
}
#endif

#endif
