/**
 *	@file NPWTest.c
 *	@brief Test program for PDF Output.
 *	@author Yves Piguet.
 *	@copyright 2008-2013, Yves Piguet.
 */

/* License: new BSD license (see NPW.h) */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "NPWBase.h"
#include "NPWStdAlloc.h"
#include "NPWStdOutput.h"
#include "NPWGraphics.h"
#include "NPWTransparency.h"
#include "NPWInfo.h"
#include "NPWText.h"
#include "NPWTextMarkup.h"

/// Set to TRUE if the markup for a horizontal rule has been decoded
static NPWBoolean drawRule;

/// Markup callback for horizontal rule (backslash-hyphen)
static NPWMarkupActionFlag markupCBHorRule(void *userData,
		NPWConstText t, NPWInt len,
		NPWInt *count,
		NPWCharStyle *style,
		NPWParFormat *parFormat,
		NPWCharStyle const *defaultStyle,
		NPWParFormat const *defaultParFormat)
{
	if (len >= 2 && t[0] == '\\' && t[1] == '-')
	{
		drawRule = TRUE;
		*count += 2;
		return kNPWMarkupActionFlagSuspend;
	}
	else
		return kNPWMarkupActionFlagNotRecognized;
}

/// Table of markup callbacks
static NPWMarkupCBEntry markupCBTable[] =
{
	{NPWMarkupCBCharStyle, NULL},
	{markupCBHorRule, NULL},
	{NULL, NULL}
};

#define WIDTH 17	///< test raster image width
#define HEIGHT 17	///< test raster image height
#define RGB 0		///< 0 if test raster image is grayscale, 1 if RGB

/// Application entry point
int main(int argc, char **argv)
{
	NPW npw;
	static unsigned char im[3 * WIDTH * HEIGHT];
	int i;
	
	NPWInit(&npw);
	NPWSetAllocCallback(&npw, NPWAllocStdlib, NULL);
	NPWSetWriteCallback(&npw, NPWWriteStdOutput, (void *)stdout);
	
	NPWBeginPage(&npw, kNPWPaperWidthA4, kNPWPaperHeightA4);
	
	NPWGraphicsRGBStroke(&npw, 1, 0, 0);
	NPWGraphicsSetLineWidth(&npw, 5);
	NPWGraphicsMove(&npw, 150, 250);
	NPWGraphicsLine(&npw, 150, 350);
	NPWGraphicsStroke(&npw);
	NPWGraphicsBeginText(&npw);
	NPWGraphicsSetTextFont(&npw, NPWFontNameTimesBoldItalic, 24);
	NPWGraphicsGray(&npw, 0.5);
	NPWGraphicsMoveText(&npw, 100, 400);
	NPWGraphicsShowText(&npw, "Hello, NPW!", -1);
	NPWGraphicsEndText(&npw);
	
	NPWGraphicsBeginText(&npw);
	NPWGraphicsSetTextFont(&npw, NPWFontNameCourier, 18);
	NPWGraphicsSetTextLeading(&npw, 20);
	NPWGraphicsMoveText(&npw, 400, 400);
	NPWGraphicsGray(&npw, 0.2);
	NPWGraphicsShowText(&npw, "1st line", -1);
	NPWGraphicsNLShowText(&npw, "2nd line", -1);
	NPWGraphicsNLShowText(&npw, "3rd line", -1);
	NPWGraphicsEndText(&npw);
	
	NPWGraphicsSave(&npw);
	NPWGraphicsConcatCTM(&npw, 50, 0, 0, 50, 250, 250);
	NPWInlineImageBegin(&npw, WIDTH, HEIGHT, RGB ? "/RGB" : "/G", 8,
			"[" NPWFilterNameASCIIHex "]");
	NPWFilterBegin(&npw, kNPWFilterASCIIHex);// | kNPWFilterRunLength);
	if (RGB)
		for (i = 0; i < WIDTH * HEIGHT; i++)
		{
			im[3 * i] = 10 * (i % WIDTH);
			im[3 * i + 1] = 10 * (i / WIDTH);
			im[3 * i + 2] = 0;
		}
	else
		for (i = 0; i < WIDTH * HEIGHT; i++)
			im[i] = 10 * (i % WIDTH);
	NPWFilterWrite(&npw, im, (RGB ? 3 : 1) * WIDTH * HEIGHT);
	NPWFilterEnd(&npw);
	NPWInlineImageEnd(&npw);
	NPWGraphicsRestore(&npw);
	
	NPWGraphicsMove(&npw, 200, 500);
	NPWGraphicsLine(&npw, 200, 460);
	NPWGraphicsSetLineWidth(&npw, 1);
	NPWGraphicsGrayStroke(&npw, 0);
	NPWGraphicsStroke(&npw);
	NPWTextWrite(&npw, "Left", -1,
		200, 500,
		NULL, kNPWAlignmentLeft);
	NPWTextWrite(&npw, "Right", -1,
		200, 480,
		NULL, kNPWAlignmentRight);
	NPWTextWrite(&npw, "Center", -1,
		200, 460,
		NULL, kNPWAlignmentCenter);
	
	NPWGraphicsRectangle(&npw, 400, 100, 30, 20);
	NPWGraphicsStroke(&npw);
	
	NPWGraphicsSave(&npw);
	NPWGraphicsRectangle(&npw, 500, 100, 30, 20);
	NPWGraphicsClip(&npw);
	NPWGraphicsEndPath(&npw);
	NPWGraphicsRectangle(&npw, 480, 90, 100, 50);
	NPWGraphicsFill(&npw);
	NPWGraphicsRestore(&npw);
	
	NPWGraphicsSave(&npw);
	NPWGraphicsRGBStroke(&npw, 1, 0, 0);
	NPWGraphicsSetLineWidth(&npw, 5);
	NPWGraphicsRectangle(&npw, 400, 200, 80, 50);
	NPWGraphicsStroke(&npw);
	NPWGraphicsSetTransparency(&npw, 0.8, 0.8, kNPWBlendModeKeyNormal);
	NPWGraphicsRGB(&npw, 0, 0, 0.8);
	NPWGraphicsRectangle(&npw, 440, 225, 80, 50);
	NPWGraphicsFill(&npw);
	NPWGraphicsRestore(&npw);
	
	NPWEndPage(&npw);
	
	NPWBeginPage(&npw, kNPWPaperWidthA4, kNPWPaperHeightA4);
	NPWGraphicsMove(&npw, 100, 100);
	NPWGraphicsLine(&npw, 400, 500);
	NPWGraphicsStroke(&npw);
	
	NPWGraphicsMove(&npw, 100, 610);
	NPWGraphicsLine(&npw, 250, 610);
	NPWGraphicsStroke(&npw);
	{
		char const *str = "\\0>\\c\\18sfoo bar baz\n\\12s\\10>\\jLorem \\iipsum\\i dolor sit amet, \\bconsectetuer\\b adipiscing elit. "
			"Aenean eget purus \\Hvitae arcu tempus sodales\\T. Sed varius nisl quis tellus. Mauris tempus "
			"tellus eget eros. Nulla enim.\n\\-"
			"Nam venenatis, nunc eu adipiscing\n\\20+commodo, ante ipsum adipiscing enim, eu convallis risus pede non odio.";
		NPWFloat lmw[] = {150};
		NPWTextLayoutState state;
		
		drawRule = FALSE;
		
		NPWTextLayoutInit(&state, markupCBTable);
		NPWTextLayoutSetPosition(&state, 100, 600);
		NPWTextLayoutSetLineWidths(&state, lmw, 1);
		NPWTextLayoutSetText(&state, str, -1);
		while (NPWTextLayoutWrite(&npw, &state) == kNPWErrSuspended && drawRule)
		{
			NPWFloat ybottom;
			
			drawRule = FALSE;
			NPWTextLayoutVerticalSkip(&state, 12, &ybottom);
			NPWGraphicsMove(&npw, 100, ybottom);
			NPWGraphicsLine(&npw, 100 + lmw[0], ybottom);
			NPWGraphicsStroke(&npw);
		}
	}
	
	NPWEndPage(&npw);
	
	NPWInfoAddEntry(&npw, kNPWInfoKeyTitle, "Test of NPW library");
	NPWInfoAddEntry(&npw, kNPWInfoKeyAuthor, "Yves Piguet");
	NPWInfoAddEntry(&npw, kNPWInfoKeyCreator, "NPW");
	
	NPWTerminate(&npw);
	
	return 0;
}
