/**
 *	@file NPWText.c
 *	@brief Helper functions for text layout
 *	@author Yves Piguet.
 *	@copyright 2007-2013, Yves Piguet.
 */

/* License: new BSD license (see header file) */

#include "NPWText.h"

#define kDefaultFont NPWFontNameCourier	///< default font name
#define kDefaultFontSize 12	///< default font size

/// Check error returned by a function and returns it unless kNPWErrOk
#define Chk(e) \
	do { \
		err = (e); \
		if (err != kNPWErrOk) \
			return err; \
	} while (0)

NPWBoolean NPWBreakNextLine(NPWConstText str, NPWInt len, NPWInt start,
		NPWFloat lineMaxWidth,
		NPWMarkupCBEntry const markupCBTable[],
		NPWInt *lineBr, NPWFloat *lineWidth, NPWInt *nSpaces,
		NPWBoolean *suspended,
		NPWCharStyle *style,
		NPWParFormat *parFormat,
		NPWCharStyle const *defaultStyle,
		NPWParFormat const *defaultParFormat)
{
	NPWInt cIx;	// index of current character in str
	NPWInt spIx;	// index of last space in str (-1 if none)
	NPWFloat w;	// current line length
	NPWInt nSp;	// number of spaces in current line
	NPWFloat w0;	// line length until spIx
	NPWFloat wch;	// width of current character
	NPWFont font;	// current font
	
	// cstring if len < 0
	if (len < 0)
		for (len = 0; str[len]; len++)
			;
	
	if (len <= start)
		return FALSE;
	
	(void)NPWFindFont(style, &font);
	
	for (cIx = start, w = 0, spIx = -1, w0 = w, nSp = 0;
			cIx < len && str[cIx] != '\n';
			)
	{
		// look for markup
		if (markupCBTable)
		{
			NPWInt i, count;
			NPWMarkupActionFlag action;
			
			for (i = count = 0; markupCBTable[i].f; i++)
			{
				action = markupCBTable[i].f(markupCBTable[i].userData,
						str + cIx, len - cIx,
						&count, style, parFormat,
						defaultStyle, defaultParFormat);
				cIx += count;
				if (!(action & kNPWMarkupActionFlagNotRecognized))
				{
					if (action & kNPWMarkupActionFlagNewStyle)
						(void)NPWFindFont(style, &font);
					if (action & kNPWMarkupActionFlagSuspend)
					{
						if (suspended)
							*suspended = TRUE;
						goto endOfPar;
					}
					if (action & kNPWMarkupActionFlagEndOfPar)
						goto endOfPar;
					// other flags don't have any effect on horizontal size
					goto contMainLoop;
				}
			}
		}
		
		if (str[cIx] == ' ')
		{
			spIx = cIx;
			w0 = w;
			nSp++;
		}
		NPWTextSize(&font, &str[cIx], 1, &wch, NULL);
		w += wch;
		if (w > lineMaxWidth
				&& w > wch	// at least 1 char
				&& lineMaxWidth >= 0)	// no line break if negative line width
		{
			*lineWidth = spIx >= 0 ? w0 : w - wch;
			*nSpaces = spIx >= 0 ? nSp - 1 : nSp;
			w = spIx >= 0 ? w - w0 : wch;
			if (spIx >= 0)	// remove space width
			{
				NPWTextSize(&font, " ", 1, &wch, NULL);
				w -= wch;
			}
			nSp = str[cIx] == ' ' ? 1 : 0;
			*lineBr = spIx >= 0 ? spIx + 1 : cIx;
			return TRUE;
		}
		
		cIx++;
contMainLoop:
		;
	}
	
endOfPar:
	*lineWidth = w;
	*nSpaces = nSp;
	*lineBr = cIx;
	
	return TRUE;
}

void NPWTextLayoutInit(NPWTextLayoutState *state,
		NPWMarkupCBEntry const markupCBTable[])
{
	state->markupCBTable = markupCBTable;
	
	state->style.fontBaseName = "Times";
	state->style.size = 12;
	state->style.isBold = FALSE;
	state->style.isItalic = FALSE;
	state->style.isUnderlined = FALSE;
	state->style.isHidden = FALSE;
	state->style.verticalAlignment = kNPWCharVertAlignmentNormal;
	state->style.color.r = 0;
	state->style.color.g = 0;
	state->style.color.b = 0;
	
	state->parFormat.alignment = kNPWAlignmentJustify;
	state->parFormat.indent0 = 24;
	state->parFormat.indent = 0;
	state->parFormat.lineSpacing = 1200;
	state->parFormat.parSpacing = 400;
	
	state->defaultStyle = state->style;
	state->defaultParFormat = state->parFormat;
	
	state->x = 100;
	state->y = 800;
	state->bottom = 0;
}

void NPWLayoutSetDefaultCharStyle(NPWTextLayoutState *state,
		NPWCharStyle const *style)
{
	state->style = *style;
}

void NPWLayoutSetDefaultParagraphStyle(NPWTextLayoutState *state,
		NPWParFormat const *parFormat)
{
	state->parFormat = *parFormat;
}

void NPWTextLayoutSetPosition(NPWTextLayoutState *state,
		NPWFloat x, NPWFloat y)
{
	state->x = x;
	state->y = y;
}

void NPWTextLayoutSetLineWidths(NPWTextLayoutState *state,
		NPWFloat const lineMaxWidths[], NPWInt lineMWLen)
{
	state->lineMaxWidths = lineMaxWidths;
	state->lineMWLen = lineMWLen;
}

void NPWTextLayoutSetBottom(NPWTextLayoutState *state,
		NPWFloat bottom)
{
	state->bottom = bottom;
}

void NPWTextLayoutSetText(NPWTextLayoutState *state,
		NPWConstText str, NPWInt len)
{
	state->str = str;
	if (len < 0)
		for (len = 0; str[len]; len++)
			;
	state->len = len;
	state->i = 0;
	state->newParagraph = TRUE;
}

NPWErr NPWTextLayoutWrite(NPW *pdf,
		NPWTextLayoutState *state)
{
	NPWInt lineBr;
	NPWFloat lineWidth;
	NPWInt nSpaces;
	NPWFloat xlast = 0, xnew = 0;
	NPWInt spanB, spanE;	// span of text str[spanB..spanE-1] with same style
	NPWInt spanBNext;	// beginning of next span (after markup)
	NPWInt currentCharAlignment = kNPWCharVertAlignmentNormal;
	NPWFloat vskip;	// vertical line skip
	NPWFloat currentIndent;
	NPWRGBColor color = {0, 0, 0};
	NPWCharStyle styleTmp;
	NPWParFormat parFormatTmp;
	NPWFloat lineMaxWidth;
	NPWBoolean suspended = FALSE;
	NPWBoolean firstMove = TRUE;	// FALSE after first NPWGraphicsMoveText
	NPWErr err;
	
	// PDF text preamble
	Chk(NPWGraphicsBeginText(pdf));
	
	// consume initial markup if any (paragraph formatting)	
nextParagraph:
	if (state->markupCBTable)
	{
		NPWInt j, count;
		NPWMarkupActionFlag action;
		
		while (state->i < state->len)
		{
			for (j = count = 0; state->markupCBTable[j].f; j++)
			{
				action = state->markupCBTable[j].f(state->markupCBTable[j].userData,
						state->str + state->i, state->len - state->i,
						&count, &state->style, &state->parFormat,
						&state->defaultStyle, &state->defaultParFormat);
				state->i += count;
				if (action & kNPWMarkupActionFlagSuspend)
				{
					Chk(NPWGraphicsEndText(pdf));
					return kNPWErrSuspended;
				}
				if (!(action & kNPWMarkupActionFlagNotRecognized))
					goto contNextMarkup;
			}
			break;
contNextMarkup:
			;
		}
	}
	
	// line spacing and text position
	Chk(NPWGraphicsSetTextFont(pdf, NPWFindFont(&state->style, NULL), state->style.size));
	Chk(NPWGraphicsSetTextLeading(pdf, state->style.size * state->parFormat.lineSpacing / 1000));
	
	xnew = state->x;
	if (firstMove)
	{
		Chk(NPWGraphicsMoveText(pdf, state->x, state->y));
		xlast = 0;
		firstMove = FALSE;
	}
	else
	{
		Chk(NPWGraphicsMoveText(pdf, state->x - xlast, firstMove ? state->y : 0));
		xlast = state->x;
	}
	
	while (state->i < state->len && state->str[state->i] != '\n' && !suspended)
	{
		// check space
		vskip = state->newParagraph
				? -state->style.size * (state->parFormat.lineSpacing + state->parFormat.parSpacing) / 1000
				: state->i > 0 ? -state->style.size * state->parFormat.lineSpacing / 1000 : 0;
		if (state->y < state->bottom)
			break;
		
		// indent for current line
		if (state->newParagraph)
			currentIndent = state->parFormat.indent + state->parFormat.indent0;
		else
			currentIndent = state->parFormat.indent;
		
		// get next line
		styleTmp = state->style;
		parFormatTmp = state->parFormat;
		lineMaxWidth
				= state->lineMWLen > 0
					? state->lineMaxWidths[state->i < state->lineMWLen ? state->i : state->lineMWLen - 1]
						- currentIndent
					: -1;
		if (!NPWBreakNextLine(state->str, state->len, state->i,
				lineMaxWidth,
				state->markupCBTable,
				&lineBr, &lineWidth, &nSpaces, &suspended,
				&styleTmp, &parFormatTmp,
				&state->defaultStyle, &state->defaultParFormat))
			break;
		
		// set spacing and go to beginning of line
		switch (state->parFormat.alignment)
		{
			case kNPWAlignmentCenter:
				xnew = (lineMaxWidth - lineWidth) / 2 + currentIndent;
				Chk(NPWGraphicsMoveText(pdf,
						xnew - xlast,
						vskip));
				xlast = xnew;
				break;
			case kNPWAlignmentRight:
				xnew = lineMaxWidth - lineWidth + currentIndent;
				Chk(NPWGraphicsMoveText(pdf,
						xnew - xlast,
						vskip));
				xlast = xnew;
				break;
			default:
				xnew = currentIndent;
				Chk(NPWGraphicsMoveText(pdf,
						xnew - xlast,
						vskip));
				xlast = xnew;
				Chk(NPWGraphicsSetWordSpace(pdf,
						lineBr < state->len && state->str[lineBr] != '\n'
									&& state->parFormat.alignment == kNPWAlignmentJustify
									&& !suspended
								? (lineMaxWidth - lineWidth) / nSpaces : 0));
				break;
		}
		state->newParagraph = FALSE;
		state->y += vskip;
		
		// break line into spans with same style
		styleTmp = state->style;
		parFormatTmp = state->parFormat;
		for (spanB = state->i; spanB < lineBr; spanB = spanBNext)
		{
			for (spanE = spanBNext = spanB; spanE < lineBr; spanE++)
				if (state->markupCBTable)
				{
					NPWInt i, count;
					NPWMarkupActionFlag action;
					
					for (i = count = 0; state->markupCBTable[i].f; i++)
					{
						action = state->markupCBTable[i].f(state->markupCBTable[i].userData,
								state->str + spanE, lineBr - spanE,
								&count, &styleTmp, &parFormatTmp,
								&state->defaultStyle, &state->defaultParFormat);
						spanBNext = spanE + count;
						if (!(action & kNPWMarkupActionFlagNotRecognized))
							goto spanDone;
					}
				}
			if (spanBNext < spanE)
				spanBNext = spanE;
spanDone:
			if (spanE > spanB)
			{
				Chk(NPWGraphicsSetTextFont(pdf, NPWFindFont(&state->style, NULL), state->style.size));
				if (state->style.color.r != color.r || state->style.color.g != color.g
						|| state->style.color.b != color.b)
				{
					color = state->style.color;
					if (color.r == color.g && color.r == color.b)
						Chk(NPWGraphicsGray(pdf, (NPWFloat)color.r / 255));
					else
						Chk(NPWGraphicsRGB(pdf,
								(NPWFloat)color.r / 255, (NPWFloat)color.g / 255, (NPWFloat)color.b / 255));
				}
				
				if (state->style.verticalAlignment != currentCharAlignment)
				{
					switch (state->style.verticalAlignment)
					{
						case kNPWCharVertAlignmentNormal:
							Chk(NPWGraphicsSetTextRise(pdf, 0));
							break;
						case kNPWCharVertAlignmentSubscript:
							Chk(NPWGraphicsSetTextRise(pdf, -styleTmp.size * 3 / 10));
							break;
						case kNPWCharVertAlignmentSuperscript:
							Chk(NPWGraphicsSetTextRise(pdf, styleTmp.size * 3 / 10));
							break;
					}
					currentCharAlignment = state->style.verticalAlignment;
				}
				if (!state->style.isHidden)
					Chk(NPWGraphicsShowText(pdf, state->str + spanB, spanE - spanB));
				else
				{
					NPWFloat w;
					NPWFont font;
					
					(void)NPWFindFont(&state->style, &font);
					NPWTextSize(&font, state->str + spanB, spanE - spanB, &w, NULL);
					Chk(NPWGraphicsHorSkip(pdf, 1000 * w / font.size));
				}
				if (state->style.isUnderlined)
				{
					// ...
				}
			}
			
			state->style = styleTmp;
			state->parFormat = parFormatTmp;
		}
		
		state->i = lineBr;
	}
	
	// paragraph break
	if (!suspended && state->i < state->len && state->str[state->i] == '\n')
	{
		state->i++;
		state->newParagraph = TRUE;
		goto nextParagraph;
	}
	
	// PDF text postamble
	Chk(NPWGraphicsEndText(pdf));
	
	return suspended ? kNPWErrSuspended : state->i < state->len ? kNPWErrIncomplete : kNPWErrOk;
}

NPWErr NPWTextLayoutVerticalSkip(NPWTextLayoutState *state,
		NPWFloat v, NPWFloat *ylow)
{
	if (state->y - v < state->bottom)
		return kNPWErrIncomplete;
	state->y -= v;
	if (ylow)
		*ylow = state->y;
	return kNPWErrOk;
}

NPWErr NPWTextWrite(NPW *pdf,
		NPWConstText str, NPWInt len,
		NPWFloat x, NPWFloat y,
		NPWCharStyle const *style, NPWAlignment alignment)
{
	NPWTextLayoutState state;
	
	NPWTextLayoutInit(&state, NULL);
	NPWTextLayoutSetPosition(&state, x, y);
	NPWTextLayoutSetLineWidths(&state, NULL, 0);
	if (style)
		state.style = *style;
	state.parFormat.alignment = alignment;
	state.parFormat.indent0 = state.parFormat.indent = 0;
	state.parFormat.lineSpacing = state.parFormat.parSpacing = 0;
	NPWTextLayoutSetText(&state, str, len);
	return NPWTextLayoutWrite(pdf, &state);
}
