/**
 *	@file NPWText.h
 *	@brief Helper functions for text layout
 *	@author Yves Piguet.
 *	@copyright 2008-2013, Yves Piguet
 */

/**
 *	@page NPWTestUsage NPW Text Layout Usage
 *
 *	Text can be displayed directly with primitive PDF commands such as BT/ET
 *	(begin/end text), Tf (set font), Td (text move to) and Tj (draw string),
 *	using NPWWrite or macros defined in NPWGraphics.h. However, achieving the
 *	desired text alignment is difficult. NPWText provides functions to get
 *	font metrics, compute line breaks and, with the help of separate text
 *	markup definitions (such as those defined in NPWTextMarkup), typeset
 *	styled text.
 *
 *	Here is the minimum amount of code to typeset text in a rectangle with
 *	justification, default font, first line indentation and space between
 *	paragraphs:
 *
 *	@code
char const *str = "foo bar baz...\nSecond paragraph";
NPWFloat lw[] = {150};  // line width (same for all lines)
NPWTextLayoutState state;
NPWErr err;

// initialize
NPWTextLayoutInit(&state, NULL);  // no markup
NPWTextLayoutSetPosition(&state, 100, 600);  // position of first line
NPWTextLayoutSetLineWidths(&state, lw, 1);  // same width for all lines
NPWTextLayoutSetText(&state, str, -1);  // null-terminated text
err = NPWTextLayoutWrite(&npw, &state);  // typeset
 *	@endcode
 *
 *	By default, NPWTextLayoutWrite starts at the position given by
 *	NPWTextLayoutSetPosition and break lines according to values given by
 *	NPWTextLayoutSetLineWidths (last value is repeated if needed) and continues
 *	until the end of the text.
 *
 *	Unless the text is known to be short, a bottom limit should be specified
 *	with NPWTextLayoutSetBottom. If NPWTextLayoutWrite exceeds it, it returns
 *	kNPWErrIncomplete; then NPWTextLayoutSetPosition, maybe
 *	NPWTextLayoutSetBottom, and NPWTextLayoutWrite should be called again,
 *	until NPWTextLayoutWrite returns a different value (usually kNPWErrOk).
 *	Typically, each NPWTextLayoutWrite is performed on a different page
 *	(bracketed between NPWBeginPage and NPWEndPage), but it is also possible
 *	to have multiple columns per page, or a picture in the middle of a page.
 *
 *	A complete PDF document can be created with the following code. Note that
 *	NPW primitives could be added to each page for headers and footers.
 *
 *	@code
char *text = ...;
int textLength = ...;
FILE *pdfFile = ...;
NPW npw;
NPWTextLayoutState state;
NPWInt margin = 25 * kNPWUnitMM;
NPWInt linewidth = kNPWPaperWidthA4 - 2 * margin;
NPWErr err;

NPWInit(&npw);
NPWSetAllocCallback(&npw, NPWAllocStdlib, NULL);
NPWSetWriteCallback(&npw, NPWWriteStdOutput, (void *)pdfFile);
NPWTextLayoutInit(&state, NULL);
NPWTextLayoutSetText(&state, text, textLength);
NPWTextLayoutSetBottom(&state, bottommargin);
NPWTextLayoutSetLineWidths(&state, &linewidth, 1);

do
{
  NPWBeginPage(&npw, kNPWPaperWidthA4, kNPWPaperHeightA4);
  NPWTextLayoutSetPosition(&state, margin, kNPWPaperHeightA4 - margin);
  err = NPWTextLayoutWrite(&npw, &state);
  NPWEndPage(&npw);
} while (err == kNPWErrIncomplete);

NPWTerminate(&npw);
 *	@endcode
 */

#ifndef __NPWText__
#define __NPWText__

#ifdef __cplusplus
extern "C" {
#endif

#include "NPWBase.h"
#include "NPWGraphics.h"
#include "NPWMetrics.h"

/// Horizontal alignment
typedef enum
{
	kNPWAlignmentLeft = 0,	///< flush left
	kNPWAlignmentCenter,	///< center
	kNPWAlignmentRight,	///< flush right
	kNPWAlignmentJustify,	///< justification
} NPWAlignment;

/// Paragraph format
typedef struct
{
	NPWAlignment alignment;	///< horizontal alignment
	NPWFloat indent0;	///< additional indenting for first line
	NPWFloat indent;	///< indenting
	NPWInt lineSpacing;	///< line spacing (1000 = font size)
	NPWInt parSpacing;	///< additional line spacing between paragraphs (1000 = 1 blank line)
} NPWParFormat;

/// Action flags of markup callbacks
typedef enum
{
	kNPWMarkupActionFlagNop = 0,	///< no action
	kNPWMarkupActionFlagNotRecognized = 1,	///< no recognized markup
	kNPWMarkupActionFlagEndOfPar = 2,	///< end of paragraph
	kNPWMarkupActionFlagNewStyle = 4,	///< new character style
	kNPWMarkupActionFlagNewParFormat = 8,	///< new paragraph format
	kNPWMarkupActionFlagSuspend = 16	///< end of par + suspend text rendering with kNPWErrSuspended
} NPWMarkupActionFlag;

/** Callback to decode text markup.
	@param[in,out] userData callback-specific data
	@param[in] t text whose beginning is analyzed
	@param[in] len length of t
	@param[out] count number of characters of markup
	@param[in,out] style character style
	@param[in,out] parFormat paragraph format
	@param[in] defaultStyle default character style
	@param[in] defaultParFormat default paragraph format
	@return kNPWMarkupActionFlagNotRecognized if no markup has been recognized, or
	flags describing the markup effect
*/
typedef NPWMarkupActionFlag (*NPWMarkupCB)(void *userData,
		NPWConstText t, NPWInt len,
		NPWInt *count,
		NPWCharStyle *style,
		NPWParFormat *parFormat,
		NPWCharStyle const *defaultStyle,
		NPWParFormat const *defaultParFormat);

/// Element of table of markup callbacks
typedef struct
{
	NPWMarkupCB f;	///< callback, or NULL for last
	void *userData;	///< application-specific value
} NPWMarkupCBEntry;

/// State of text layout
typedef struct
{
	NPWMarkupCBEntry const *markupCBTable;	///< table of markup callbacks 
	NPWCharStyle style;	///< current character style
	NPWParFormat parFormat;	///< current paragraph format
	NPWCharStyle defaultStyle;	///< default character style
	NPWParFormat defaultParFormat;	///< default paragraph format
	NPWFloat x;	///< left margin
	NPWFloat y;	///< base line of next line
	NPWFloat const *lineMaxWidths;	///< array of maximum width of lines (last value is repeated)
	NPWInt lineMWLen;	///< number of elements in lineMaxWidths
	NPWFloat bottom;	///< lower baseline limit along y axis
	NPWConstText str;	///< text
	NPWInt len;	///< length of text
	NPWInt i;	///< offset in str
	NPWBoolean newParagraph;	///< TRUE if next line is the first line of a paragraph
} NPWTextLayoutState;

/**	Find next line break and compute length and number of spaces.
	@param[in] str text
	@param[in] len length of text (c-string if < 0); stops at first linefeed
	@param[in] start offset of first character in str
	@param[in] lineMaxWidth max width of line;
	no line break if negative, just compute line width and spaces for alignment
	@param[in] markupCBTable table of markup callbacks
	@param[out] lineBr offset of next line break
	@param[out] lineWidth width of next line (without trailing spaces)
	@param[out] nSpaces number of spaces (excluding trailing space)
	@param[in,out] suspended set to TRUE if a markup callback has returned
	kNPWMarkupActionFlagSuspend
	@param[in,out] style character style
	@param[in,out] parFormat paragraph format
	@param[in] defaultStyle default character style
	@param[in] defaultParFormat default paragraph format
	@return TRUE if next line is found, FALSE if end-of-text
*/
NPWBoolean NPWBreakNextLine(NPWConstText str, NPWInt len, NPWInt start,
		NPWFloat lineMaxWidth,
		NPWMarkupCBEntry const markupCBTable[],
		NPWInt *lineBr, NPWFloat *lineWidth, NPWInt *nSpaces,
		NPWBoolean *suspended,
		NPWCharStyle *style,
		NPWParFormat *parFormat,
		NPWCharStyle const *defaultStyle,
		NPWParFormat const *defaultParFormat);

/**	Initialize text layout state.
	@param[out] state text layout state
	@param[in] markupCBTable table of markup callbacks, or NULL if none
*/
void NPWTextLayoutInit(NPWTextLayoutState *state,
		NPWMarkupCBEntry const markupCBTable[]);

/**	Set default character style in text layout state.
	@param[in,out] state text layout state
	@param[in] style default character style
*/
void NPWLayoutSetDefaultCharStyle(NPWTextLayoutState *state,
		NPWCharStyle const *style);

/**	Set default paragraph format in text layout state.
	@param[in,out] state text layout state
	@param[in] parFormat default paragraph style
*/
void NPWLayoutSetDefaultParagraphStyle(NPWTextLayoutState *state,
		NPWParFormat const *parFormat);

/**	Set position of next line of text in text layout state.
	@param[in,out] state text layout state
	@param[in] x left margin
	@param[in] y base line of first line
*/
void NPWTextLayoutSetPosition(NPWTextLayoutState *state,
		NPWFloat x, NPWFloat y);

/**	Set array of maximum line widths.
	@param[in,out] state text layout state
	@param[in] lmw array of line maximum widths
	@param[in] n number of elements in lmw (last element is replicated)
*/
void NPWTextLayoutSetLineWidths(NPWTextLayoutState *state,
		NPWFloat const lmw[], NPWInt n);

/**	Set lower limit of text along y axis.
	@param[in,out] state text layout state
	@param[in] bottom lower limit of text
*/
void NPWTextLayoutSetBottom(NPWTextLayoutState *state,
		NPWFloat bottom);

/**	Set text.
	@param[in,out] state text layout state
	@param[in] str text
	@param[in] len text length
*/
void NPWTextLayoutSetText(NPWTextLayoutState *state,
		NPWConstText str, NPWInt len);

/**	Write formatted text.
	@param pdf reference to NPW structure
	@param[in,out] state text layout state
	@return kNPWErrOk for success, kNPWErrIncomplete if more text to write
	(y exceeds bottom), else error code
*/
NPWErr NPWTextLayoutWrite(NPW *pdf,
		NPWTextLayoutState *state);

/**	Reserve some space in text layout.
	@param[in,out] state text layout state
	@param[in] v vertical distance to skip
	@param[out] ylow y coordinate of free space bottom (base line one line above
	where next line will be displayed; free space top is the base line of previous line;
	not returned in NULL)
	@return kNPWErrOk for success, kNPWErrIncomplete if not enough space
	(y exceeds bottom), else error code
*/
NPWErr NPWTextLayoutVerticalSkip(NPWTextLayoutState *state,
		NPWFloat v, NPWFloat *ylow);

/**	Write simple styled text.
	@param pdf reference to NPW structure
	@param[in] str text
	@param[in] len length of text, or -1 if str is null-terminated
	@param[in] x position along x axis
	@param[in] y position along y axis
	@param[in] style character style
	@param[in] alignment horizontal alignment
	(kNPWAlignmentLeft, kNPWAlignmentCenter, or kNPWAlignmentRight)
	@return kNPWErrOk for success, else error code
*/
NPWErr NPWTextWrite(NPW *pdf,
		NPWConstText str, NPWInt len,
		NPWFloat x, NPWFloat y,
		NPWCharStyle const *style, NPWAlignment alignment);

#ifdef __cplusplus
}
#endif

#endif
