/**
 *	@file NPWTextMarkup.c
 *	@brief Simple text markup for PDF Output.
 *	@author Yves Piguet.
 *	@copyright 2008-2013, Yves Piguet.
 */

/* License: new BSD license (see NPW.h) */

#include "NPWTextMarkup.h"

NPWMarkupActionFlag NPWMarkupCBCharStyle(void *userData,
		NPWConstText t, NPWInt len,
		NPWInt *count,
		NPWCharStyle *style,
		NPWParFormat *parFormat,
		NPWCharStyle const *defaultStyle,
		NPWParFormat const *defaultParFormat)
{
	NPWInt n;
	NPWBoolean neg, reset;
	(void)userData;
	
	if (len >= 2 && t[0] == '\\')
	{
		*count = 1;
		n = 0;
		neg = t[*count] == '-';
		if (neg)
			(*count)++;
		for (; t[*count] >= '0' && t[*count] <= '9'; (*count)++)
			n = 10 * n + t[*count] - '0';
		if (neg)
			n = -n;
		reset = neg && n == 0;
		
		switch (t[*count])
		{
			case '\\':	// escaped backslash
				return kNPWMarkupActionFlagNop;
			case 'b':	// bold
				(*count)++;
				style->isBold = !reset && !style->isBold;
				return kNPWMarkupActionFlagNewStyle;
			case 'i':	// italic
				(*count)++;
				style->isItalic = !reset && !style->isItalic;
				return kNPWMarkupActionFlagNewStyle;
			case 'u':	// underline
				(*count)++;
				style->isUnderlined = !reset && !style->isUnderlined;
				return kNPWMarkupActionFlagNewStyle;
			case 'h':	// hidden
				(*count)++;
				style->isHidden = !reset && !style->isHidden;
				return kNPWMarkupActionFlagNewStyle;
			case '^':	// superscript
				(*count)++;
				style->verticalAlignment
						= reset ? kNPWCharVertAlignmentNormal
							: style->verticalAlignment == kNPWCharVertAlignmentSuperscript
								? kNPWCharVertAlignmentNormal : kNPWCharVertAlignmentSuperscript;
				return kNPWMarkupActionFlagNewStyle;
			case '_':	// subscript
				(*count)++;
				style->verticalAlignment
						= reset ? kNPWCharVertAlignmentNormal
							: style->verticalAlignment == kNPWCharVertAlignmentSubscript
								? kNPWCharVertAlignmentNormal : kNPWCharVertAlignmentSubscript;
				return kNPWMarkupActionFlagNewStyle;
			case 'T':	// Times
				(*count)++;
				style->fontBaseName = "Times";
				return kNPWMarkupActionFlagNewStyle;
			case 'H':	// Helvetica
				(*count)++;
				style->fontBaseName = "Helvetica";
				return kNPWMarkupActionFlagNewStyle;
			case 'C':	// Courier
				(*count)++;
				style->fontBaseName = "Courier";
				return kNPWMarkupActionFlagNewStyle;
			case 'S':	// Symbol
				(*count)++;
				style->fontBaseName = "Symbol";
				return kNPWMarkupActionFlagNewStyle;
			case 'Z':	// Zapf Dingbats
				(*count)++;
				style->fontBaseName = "ZapfDingbats";
				return kNPWMarkupActionFlagNewStyle;
			case 's':	// size
				(*count)++;
				style->size = reset ? defaultStyle->size : n;
				return kNPWMarkupActionFlagNewStyle;
			case 'r':	// right
				(*count)++;
				parFormat->alignment = kNPWAlignmentRight;
				return kNPWMarkupActionFlagNewParFormat;
			case 'l':	// left
				(*count)++;
				parFormat->alignment = kNPWAlignmentLeft;
				return kNPWMarkupActionFlagNewParFormat;
			case 'c':	// center
				(*count)++;
				parFormat->alignment = kNPWAlignmentCenter;
				return kNPWMarkupActionFlagNewParFormat;
			case 'j':	// justify
				(*count)++;
				parFormat->alignment = kNPWAlignmentJustify;
				return kNPWMarkupActionFlagNewParFormat;
			case '>':	// indent first line
				(*count)++;
				parFormat->indent0 = reset ? defaultParFormat->indent0 : n;
				return kNPWMarkupActionFlagNewParFormat;
			case '+':	// indent all lines
				(*count)++;
				parFormat->indent = reset ? defaultParFormat->indent : n;
				return kNPWMarkupActionFlagNewParFormat;
			case 'L':	// line spacing (1000 = font size)
				(*count)++;
				parFormat->lineSpacing = reset ? defaultParFormat->lineSpacing : n;
				return kNPWMarkupActionFlagNewParFormat;
			case 'K':	// additional line spacing between paragraphs (1000 = font size)
				(*count)++;
				parFormat->parSpacing = reset ? defaultParFormat->parSpacing : n;
				return kNPWMarkupActionFlagNewParFormat;
			default:
				*count = 0;
				return kNPWMarkupActionFlagNotRecognized;
		}
	}
	
	return kNPWMarkupActionFlagNotRecognized;
}
