/**
 *	@file NPWTransparency.c
 *	@brief Transparency-related functions for graphics
 *	@author Yves Piguet.
 *	@copyright 2011-2013, Yves Piguet.
 */

/* License: new BSD license (see header file) */

#include "NPWTransparency.h"
#include <string.h>

/// Check error returned by a function and returns it unless kNPWErrOk
#define Chk(e) \
	do { \
		err = (e); \
		if (err != kNPWErrOk) \
			return err; \
	} while (0)

typedef struct
{
	NPWFloat strokingAlpha;
	NPWFloat nonStrokingAlpha;
	NPWChar blendMode;	// null-terminated string
} Transparency;

static NPWConstText kDataKeyTransparency = "transparency";

/**	Write transparency resources (installed with NPWAddResourceCB).
	@param pdf reference to NPW structure
	@param userData callback user data (not used)
	@return error code (kNPWErrOk for success)
*/
static NPWErr writeTransparencyRsrc(NPW *pdf, void *userData)
{
	Transparency *t;
	NPWInt i;
	NPWBoolean hasBegun = FALSE;
	NPWErr err;
	(void)userData;
	
	t = NULL;
	for (i = 1; ; i++)
	{
		if (NPWDataGetNext(pdf, kDataKeyTransparency, (void **)&t) != kNPWErrOk)
			break;
		
		if (i == 1)
		{
			// at least one object: write header
			Chk(NPWWrite(pdf, "/ExtGState\n<<\n", -1));
			hasBegun = TRUE;
		}
		
		// write /ExtGState object
		Chk(NPWWrite(pdf, "/GS", -1));
		Chk(NPWWriteInt(pdf, i, -1));
		Chk(NPWWrite(pdf, "\n<<\n/Type /ExtGState\n/ca ", -1));
		Chk(NPWWriteFloat(pdf, t->nonStrokingAlpha));
		Chk(NPWWrite(pdf, "\n/CA ", -1));
		Chk(NPWWriteFloat(pdf, t->strokingAlpha));
		Chk(NPWWrite(pdf, "\n/BM /", -1));
		Chk(NPWWrite(pdf, &t->blendMode, -1));
		Chk(NPWWrite(pdf, "\n>>\n", -1));
	}
	if (hasBegun)
		Chk(NPWWrite(pdf, ">>\n", -1));
	
	return kNPWErrOk;
}

NPWErr NPWGraphicsSetTransparency(NPW *pdf,
		NPWFloat strokingAlpha, NPWFloat nonStrokingAlpha,
		NPWConstText blendMode)
{
	Transparency *t;
	NPWBoolean first = TRUE;
	NPWInt i;
	NPWErr err;
	
	// check if the same settings already exist
	t = NULL;
	for (i = 0; ; i++)
	{
		if (NPWDataGetNext(pdf, kDataKeyTransparency, (void **)&t) != kNPWErrOk)
		{
			// not found; add
			Chk(NPWDataAlloc(pdf, kDataKeyTransparency,
					sizeof(Transparency) + strlen(blendMode), (void **)&t));
			t->strokingAlpha = strokingAlpha;
			t->nonStrokingAlpha = nonStrokingAlpha;
			strcpy(&t->blendMode, blendMode);
			i = NPWDataCount(pdf, kDataKeyTransparency);
			break;
		}
		
		first = FALSE;
		
		if (t->strokingAlpha == strokingAlpha
				&& t->nonStrokingAlpha == nonStrokingAlpha
				&& !strcmp(&t->blendMode, blendMode))
		{
			i++;
			break;
		}
	}
	
	Chk(NPWWrite(pdf, "/GS", -1));
	Chk(NPWWriteInt(pdf, i, -1));
	Chk(NPWWrite(pdf, " gs\n", -1));
	
	if (first)
		Chk(NPWAddResourceCB(pdf, writeTransparencyRsrc, NULL));
	
	return kNPWErrOk;
}
